(function() {
  "use strict";

  Snake.Crypto.Bush = (function() {
    /**
     * Bush.
     * @class Snake.Crypto.Bush
     * @returns {Snake.Crypto.Bush} the tree.
     */
    function Bush() {
      /**
       * Order of the tree.
       * @type {Integer}
       * @memberOf Snake.Crypto.Bush#
       */
      this.order = Snake.Config.Group.BushOrder;

      /**
       * Maximum number of keys of a node.
       * @type {Integer}
       * @memberOf Snake.Crypto.Bush#
       */
      this.maxChildren = this.order;
      /**
       * Minimum number of keys of a leaf node.
       * @type {Integer}
       * @memberOf Snake.Crypto.Bush#
       */
      this.minChildren = Math.ceil(this.order / 2);

      /**
       * Root of the tree.
       * @type {Object}
       * @memberOf Snake.Crypto.Bush#
       */
      this.root = null;

      /**
       * Array of nodes that are in the special path
       * @type {Object[]}
       * @memberOf Snake.Crypto.Bush#
       */
      this.sp = [];

      /**
       * Array of internal nodes that
       * @type {Object[]}
       * @memberOf Snake.Crypto.Bush#
       */
      this.nonfullnodes = [];

      /**
       * Height of the tree.
       * @type {Integer}
       * @memberOf Snake.Crypto.Bush#
       */
      this.height = 1;

      /**
       * Array of external nodes.
       * @type Object
       * @memberOf Snake.Crypto.Bush#
       */
      this.externalNodes = {};

      /**
       * Counter used to assign an unique identifier to each key in the tree.
       * @type Integer
       * @memberOf Snake.Crypto.Bush#
       */
      this.keyCount = 0;

      /**
       * Initialization vector used to encrypt the rekying messages.
       * @type {ArrayBufferView}
       */
      this.iv = null;
    }

    /**
     * Initializes the tree.
     * @returns {Promise} resolved when the tree has been initialized.
     */
    Bush.prototype.initBush = function() {
      // creates the root node.
      this.root = this._createNode();
      // the new root is a non full node.
      this.nonfullnodes.push(this.root);
      // sets the initialization vector for the rekying message.
      this.iv = window.crypto.getRandomValues(new Uint8Array(16));
      // returns the promise of the root node rekey
      return this._rekeyNode(this.root);
    };

    /**
     * Creates a new node of the tree.
     * @param {Boolean} external true if the node is an external node, false if
     *                  the node is internal.
     * @returns {Object} the node created.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._createNode = function(external) {
      if (external) {
        return {
          /**
           * Parent node.
           * @type {Object}
           */
          parent: null,
          /**
           * The node id (which is equal to the user id associated to the node).
           * @type {Integer}
           */
          nodeId: null,
          /**
           * Cryptographic key of the node.
           * @type {{id: Integer, key: String}}
           */
          nodeKey: null
        };
      } else {
        return {
          /**
           * Parent node.
           * @type {Object}
           */
          parent: null,
          /**
           * Array of child nodes.
           * @type {Object[]}
           */
          childNodes: [],
          /**
           * Cryptographic key of the node.
           * @type {{id: Integer, key: String}}
           */
          nodeKey: null
        };
      }
    };

    /**
     * Determines whether a node is leaf or not.
     * @param {Object} node the node.
     * @returns {Boolean} true if the node is leaf, false otherwise.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._isLeaf = function(node) {
      if(node.childNodes) {
        return false;
      } else {
        return true;
      }
    };

    /**
     * Generates a new key for a node.
     * @returns {Promise<Object>} the generated node key.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._generateKey = function() {
      return Snake.Crypto.Common.generateSymmetricKey().then(function(key) {
        return {
          id: this.keyCount++,
          key: key
        };
      }.bind(this));
    };

    /**
     * Generates a new key for the node.
     * @param {Object} node the node to rekey;
     * @returns {Promise} resolved when the node has been rekeyed.
     * @memberOf BushInternalNode#
     * @private
     */
    Bush.prototype._rekeyNode = function(node) {
      return this._generateKey().then(function(newKey) {
        node.nodeKey = newKey;
      });
    };

    /**
     * Returns the level of a node.
     * @param {Object} node the node.
     * @returns {Integer} the level of the node.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._nodeLevel = function(node) {
      var level = 0;

      while (node.parent) {
        level++;
        node = node.parent;
      }

      return level;
    };

    /**
     * Returns an internal node of the tree where to insert a new external node.
     * @returns {Object} an internal node.
     * @memberOf Snake.Crypto.Bush#
     */
    Bush.prototype._getInsertPoint = function() {
      // if the special path is not empty take the internal node that has the
      // largest level
      if (this.sp.length !== 0) {
        // TODO there is only one special path in the tree, picking up the latest
        //      node in array should be ok
        var index = 0;
        for (var i = 1; i < this.sp.length; i++) {
          if (this._nodeLevel(this.sp[i]) > this._nodeLevel(this.sp[index])) {
            index = i;
          }
        }

        // the node remains in the special path if it as at most Bush.minChildren-1
        // otherwise it's moved into the nonfull internal nodes array
        var node = this.sp[index];
        if (node.childNodes.length === (this.minChildren - 1)) {
          this.sp.splice(index, 1);
          this.nonfullnodes.push(node);
        }

        return node;
      }

      // returns a nonfull internal node not in the special path
      if (this.nonfullnodes.length === 0) {
        return null;
      } else {
        var index = 0;
        for (var i = 1; i < this.nonfullnodes.length; i++) {
          if (this.nonfullnodes[i].childNodes.length >
              this.nonfullnodes[index].childNodes.length) {
            index = i;
          }
        }

        // if the node will become full after the insertion remove from the list
        // of nonfull nodes
        var node = this.nonfullnodes[index];
        if (node.childNodes.length === (this.maxChildren - 1)) {
          this.nonfullnodes.splice(index, 1);
        }

        return node;
      }
    };

    /**
     * Returns the externa node of a particular user.
     * @param {Interger} nodeId the node id.
     * @returns {Object} the external node of the user.
     * @private
     */
    Bush.prototype._getNode = function(nodeId) {
      return this.externalNodes[nodeId];
    };

    /**
     * Encrypts the key of a node.
     * @param {Object} node the node to encrypt.
     * @param {String} key the key to be used to encrypt.
     * @returns {Promise<Object>} the rekeying message.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._encryptNodeKey = function(node, key) {
      var oldKeyId = node.nodeKey.id;
      var rawKey = node.nodeKey.key;
      return Snake.Crypto.Common.symmetricEncryption(key.key, rawKey, this.iv)
          .then(function(encryptedData) {
        return {
          data: encryptedData,
          encKey: oldKeyId,
          withKey: key.id
        };
      }.bind(this));
    };

    /**
     * Performs a rekey starting from a node going up to an other node.
     * The new key of nodes is encrypted using the keys of the children of the
     * node.
     * @param {Object} node the node where to start the rekey.
     * @param {Object} stopNode the node where to stop the rekey.
     * @returns {Promise<Object>} the rekeying messages.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._rekeyNodesWithSubnodesKeys = function(node, stopNode) {
      var rekey = [];

      var operation = function() {
        if (!node || node === stopNode) {
          return Snake.identityPromise();
        }

        var promise = this._rekeyNode(node).then(function() {
          var nodePromises = [];
          for (var i = 0; i < node.childNodes.length; i++) {
            nodePromises.push(this._encryptNodeKey(node, node.childNodes[i].nodeKey));
          }

          return Promise.every.apply(this, nodePromises).then(function(res) {
            rekey.push(res);
            node = node.parent;
          });
        }.bind(this));

        return promise.then(operation);
      }.bind(this);

      return operation().then(function() {
        return rekey;
      });
    };

    /**
     * Performs a rekey starting from a node going up to an other node.
     * The new key of nodes is encrypted using the old key node.
     * @param {Object} node the node where to start the rekey.
     * @param {Object} stopNode the node where to stop the rekey.
     * @param {Object} key the key used to encrypt the new node key, if is not
     *                 defined the old node key will be used.
     * @param {Boolean} skipRekey true to skip the rekey operation.
     * @returns {Promise<Object>} the rekeying messages.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._rekeyNodesWithNodeKey = function(node, stopNode, key, skipRekey) {
      var oldKeys = [];
      var rekeyPromises = [];

      var tmpNode = node;
      while (tmpNode && tmpNode !== stopNode) {
        oldKeys.push(tmpNode.nodeKey);

        if (!skipRekey) {
          rekeyPromises.push(this._rekeyNode(tmpNode));
        } else {
          rekeyPromises = [Snake.identityPromise()];
        }

        tmpNode = tmpNode.parent;
      }

      if (rekeyPromises.length === 0) {
        return Snake.identityPromise([]);
      }

      return Promise.every.apply(this, rekeyPromises)
          .then(function() {
        var nodePromises = [];

        var tmpNode = node;
        while (tmpNode && tmpNode !== stopNode) {
          if (key) {
            nodePromises.push(this._encryptNodeKey(tmpNode, key));
          } else if (oldKeys[0]) {
            nodePromises.push(this._encryptNodeKey(tmpNode, oldKeys.shift()));
          }

          tmpNode = tmpNode.parent;
        }

        if (nodePromises.length === 0) {
          return Snake.identityPromise([]);
        } else {
          return Promise.every.apply(this, nodePromises);
        }
      }.bind(this));
    };

    /**
     * Performs a rekey on two nodes
     * @param {Object} nodeA
     * @param {Object} nodeB
     * @returns {Promise<Object>} the rekying message.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._rekeyRedistributedNodes = function(nodeA, nodeB) {
      var promiseNodeA = this._rekeyNodesWithSubnodesKeys(nodeA, nodeA.parent);
      var promiseNodeB = this._rekeyNodesWithSubnodesKeys(nodeB, nodeB.parent);

      return Promise.every.apply(this, [promiseNodeA, promiseNodeB]);
    };

    /**
     * Performs a rekeying on two node that will be merged.
     * @param {Object} otherNode the node that will be
     * @param {Object} intoNode the node
     * @returns {Promise<Object[]>} the rekeying message.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._rekeyMergedNodes = function(otherNode, intoNode) {
      // XXX due to the asynchronous nature of crypto operations this
      //     function must copy the value of the nodes, otherwise theis values
      //     will change because of the merge operation.
      var copyNode = function(node) {
        return {
          parent: node.parent,
          childNodes: node.childNodes.slice(0),
          nodeKey: node.nodeKey
        };
      };

      var copyOtherNode = copyNode(otherNode);
      var copyIntoNode = copyNode(intoNode);

      var result = [];
      return this._rekeyNodesWithSubnodesKeys(copyIntoNode, copyIntoNode.parent)
      .then(function(res) {
        result.push(res);
        return this._encryptNodeKey(copyIntoNode, copyOtherNode.nodeKey);
      }.bind(this)).then(function(res) {
        result.push(res);
        intoNode.nodeKey = copyIntoNode.nodeKey;
        return result;
      }.bind(this));
    };

    /**
     * Inserts an user into the tree.
     * @param {Integer} userId the id of the user to insert.
     * @returns {Promise} the rekeying message.
     * @memberOf Snake.Crypto.Bush#
     */
    Bush.prototype.insert = function(userId) {
      //
      var oldRoot = this.root;
      var newRoot = null;

      // if the user is already present in the tree do nothing
      var userNode = this._getNode(userId);
      if (userNode) {
        return Snake.identityPromise();
      }

      // creates the external node and inserts it into the list of external nodes
      var z = this._createNode(true);
      z.nodeId = userId;
      this.externalNodes[userId] = z;

      // gets a node where to insert the new user
      var y = this._getInsertPoint();

      // if there aren't nodes where to add new user creates a new root and
      // attaches the old tree as its children
      if (!y) {
        y = this._createNode();

        // adds the new root to the nonfull nodes array
        this.nonfullnodes.push(y);

        // the current root becomes a child of y
        y.childNodes.push(this.root);
        this.root.parent = y;
        this.root = y;

        // the height of the tree grows
        this.height++;

        //
        newRoot = y;
      }

      // creates a chain of internal nodes in order to place the external node
      // at the correct level
      var numInternalNodes = this.height - this._nodeLevel(y) - 1;
      var chain = [];
      if (numInternalNodes !== 0) {
        var node = this._createNode();
        chain.push(node);
      }
      for (var i = 1; i < numInternalNodes; i++) {
        var node = this._createNode();
        chain.push(node);
        chain[i - 1].childNodes.push(node);
        chain[i].parent = chain[i - 1];
      }

      // the chain becomes the tail of the current sp
      this.sp = this.sp.concat(chain);

      // attaches z to the chain of nodes
      chain.push(z);
      if (numInternalNodes > 0) {
        chain[numInternalNodes - 1].childNodes.push(z);
        z.parent = chain[numInternalNodes - 1];
      }

      // the root of the chain becomes a child of y
      y.childNodes.push(chain[0]);
      chain[0].parent = y;

      return this._insertRekey(y, chain, oldRoot, newRoot);
    };

    /**
     * Creates the rekeying message for an insertion operation.
     * @param {Object} insertionNode
     * @param {Object[]} chain
     * @param {Object} oldRoot
     * @param {Object} newRoot
     * @returns {Promise} the rekeying message.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._insertRekey = function(insertionNode, chain, oldRoot, newRoot) {
      var oldRootKey = oldRoot.nodeKey;

      var rekeyMessage = {
        other: {
          iv: this.iv
        },
        newUser: {},
        oldRootKey: oldRoot.nodeKey.key
      };

      var userNode = chain[chain.length - 1];

      // assigns the node key to the external node of the new user
      return this._rekeyNode(userNode).then(function() {
      // performs a rekey on the path from the new user node to the insertio node
        return this._rekeyNodesWithNodeKey(userNode.parent, insertionNode, null);
      }.bind(this)).then(function() {

        // if a new root has been created then rekey that node with the old root
        // otherwise the group key is not changed
        if (newRoot) {
          return this._rekeyNodesWithNodeKey(newRoot, null, oldRootKey);
        } else {
          return Snake.identityPromise([false]);
        }
      }.bind(this)).then(function(res) {
        rekeyMessage.other.rootKey = res[0];

        // collects all the keys that are needed by the new user
        rekeyMessage.newUser.auxiliaryKeys = [];
        var tmpNode = userNode;
        while (tmpNode !== this.root) {
          var nodeKey = _.clone(tmpNode.nodeKey);
          rekeyMessage.newUser.auxiliaryKeys.push(nodeKey);
          tmpNode = tmpNode.parent;
        }
        var rootKey = _.clone(this.root.nodeKey);
        rekeyMessage.newUser.rootKey = rootKey;

        // packs every rekey message into a single object
        return Snake.identityPromise(rekeyMessage);
      }.bind(this));
    };

    /**
     * Creates the rekeying message for a deletion operation.
     * @param {Object} changedNodes
     * @returns {Promise} the rekeying message.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._removeRekey = function(changedNodes) {
      var rekeyMessage = {
        keys: [],
        iv: this.iv,
        oldRootKey: changedNodes.oldRootKey
      };
      var rekeyPromise;

      if (changedNodes.rootMerged) {
      // the height of the tree has decreased by one
        if (changedNodes.otherNodes.length === 0) {
        // special case when it is removed a node in the special path
          rekeyPromise = this._rekeyNodesWithSubnodesKeys(this.root, null)
          .then(function(res) {
            rekeyMessage.keys = res;
            return rekeyMessage;
          }.bind(this));
        } else {
        // normal root merge
          rekeyMessage.keys = Array.prototype.concat.apply([], changedNodes.otherNodes);
          rekeyPromise = Snake.identityPromise(rekeyMessage);
        }
      } else {
      // the height of the tree is not changed
        rekeyMessage.keys = Array.prototype.concat.apply([], changedNodes.otherNodes);

        rekeyPromise = this._rekeyNodesWithSubnodesKeys(changedNodes.rekeyNode, this.root)
        .then(function(res) {
          rekeyMessage.keys = rekeyMessage.keys.concat(
              Array.prototype.concat.apply([], res));
          return this._rekeyNodesWithSubnodesKeys(this.root);
        }.bind(this)).then(function(res) {
          rekeyMessage.keys = rekeyMessage.keys.concat(
              Array.prototype.concat.apply([], res));
          return Snake.identityPromise(rekeyMessage);
        }.bind(this));
      }

      return rekeyPromise;
    };

    /**
     * Removes an user from the tree.
     * @param {Integer} userId the id of the user to remove.
     * @returns {Promise<Object[]>} the rekying message.
     * @memberOf Snake.Crypto.Bush#
     */
    Bush.prototype.remove = function(userId) {
      // gets the node that corresponds to the user
      var z = this._getNode(userId);
      if (!z) {
        return;
      }

      var changedNodes = {
        rekeyNode: z.parent,
        otherNodes: [],
        rootMerged: false,
        oldRootKey: this.root.nodeKey.key
      };

      // removes the node from the tree
      delete this.externalNodes[userId];

      // removes the node from the parent of z
      var pz = z.parent;
      pz.childNodes.splice(pz.childNodes.indexOf(z), 1);

      // the parent of z now is nonfull, if is not already inside the nonfull
      // internal node array insert it
      if (pz.childNodes.length > 0 && this.nonfullnodes.indexOf(pz) === -1) {
        this.nonfullnodes.push(pz);
      }

      // move one level up
      z = pz;

      var fixTree = function() {
        //
        if(z === this.root ||
          (z.childNodes.length !== (this.minChildren-1) && z.childNodes.length !== 0)) {
          return Snake.identityPromise();
        }

        var promise;

        pz = z.parent;
        var indexSp = this.sp.indexOf(z);
        if (indexSp !== -1) {
          // the node is in the special path
          // remove z from the sp
          this.sp.splice(indexSp, 1);
          // the parent of z has no child left after that z is removed
          pz.childNodes.splice(pz.childNodes.indexOf(z), 1);

          // move up one level
          z = pz;
          changedNodes.rekeyNode = z;
          promise = Snake.identityPromise();
        } else {
          // the node is not in the special path
          // tries to redistribute one element from a sibling of z
          var sz = this._richSibling(z);
          if (sz) {
            // moves a child from sz to z
            var elem = sz.childNodes.shift();
            elem.parent = z;
            z.childNodes.push(elem);

            // sz now is nonfull, if is not already inside the nonfull internal
            // node array insert it
            if (this.nonfullnodes.indexOf(sz) === -1) {
              this.nonfullnodes.push(sz);
            }

            //
            promise = this._rekeyRedistributedNodes(sz, z).then(function(res) {
              changedNodes.otherNodes = Array.prototype.concat.apply([], res);
            });

            changedNodes.rekeyNode = changedNodes.rekeyNode.parent;

            z = this.root;
            return promise;
          }

          if (pz.childNodes.length > 1) {
            // merges z with a sibling of z
            for (var i = 0; i < pz.childNodes.length; i++) {
              var node = pz.childNodes[i];
              if (node !== z) {
                //
                promise = this._rekeyMergedNodes(node, z).then(function(res) {
                  changedNodes.otherNodes = changedNodes.otherNodes.concat(
                      Array.prototype.concat.apply([], res)
                  );
                });

                changedNodes.rekeyNode = pz;

                for (var j = 0; j < node.childNodes.length; j++) {
                  node.childNodes[j].parent = z;
                }
                z.childNodes = z.childNodes.concat(node.childNodes);
                pz.childNodes.splice(pz.childNodes.indexOf(node), 1);

                // node will be deleted, remove it from the non full internal node
                // array and from sp (if present)
                this.nonfullnodes.splice(this.nonfullnodes.indexOf(node), 1);
                indexSp = this.sp.indexOf(node);
                if (indexSp !== -1) {
                  this.sp.splice(indexSp, 1);
                }

                // if z is now full remove it from the non full internal node array
                if (z.childNodes.length === this.maxChildren) {
                  this.nonfullnodes.splice(this.nonfullnodes.indexOf(z), 1);
                }

                // the parent of z (and node) now is nonfull, insert it into the
                // non full internal node array if is not already present
                if (this.nonfullnodes.indexOf(pz) === -1) {
                  this.nonfullnodes.push(pz);
                }

                break;
              }
            }

            z = pz;
          } else {
            // adds z to sp and removes from nonfull internal node array
            this.sp.push(z);
            this.nonfullnodes.splice(this.nonfullnodes.indexOf(z), 1);

            z = this.root;
            return;
          }
        }

        //
        return promise.then(fixTree);
      }.bind(this);

      return fixTree().then(function() {
        // the height of the tree is decreased by one
        if (z === this.root && z.childNodes.length < 2 && this.height > 1) {
          this.height--;

          changedNodes.rootMerged = true;

          // the root will be deleted, it must be also deleted from the array of non
          // full internal nodes
          this.nonfullnodes.splice(this.nonfullnodes.indexOf(this.root), 1);

          // the new root is the only child of z
          this.root = z.childNodes.pop();
          this.root.parent = null;

          // the root of the tree can't be in the sp
          var indexSp = this.sp.indexOf(this.root);
          if (indexSp !== -1) {
            this.sp.splice(indexSp, 1);
          }
        }

        return this._removeRekey(changedNodes);
      }.bind(this));
    };

    /**
     *
     * @param {Object} z
     * @returns {Object}
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._richSibling = function(z) {
      var pz = z.parent;

      for (var i = 0; i < pz.childNodes.length; i++) {
        var child = pz.childNodes[i];

        if (child !== z) {
          if (this.sp.indexOf(child) !== -1 && child.childNodes.length > 1) {
            return child;
          }
        }
      }

      for (var i = 0; i < pz.childNodes.length; i++) {
        var child = pz.childNodes[i];

        if (child !== z) {
          if (this.sp.indexOf(child) === -1 &&
              child.childNodes.length > (this.minChildren)) {
            return child;
          }
        }
      }
    };

    /**
     * Validates the tree.
     * @throws {String} when the tree violates a property of the Bush tree.
     * @memberOf Snake.Crypto.Bush#
     */
    Bush.prototype.validate = function() {
      for (var i = 0; i < this.sp.length; i++) {
        if (this.sp[i].length > this.minChildren - 1) {
          throw "node in the sp has too many children";
        }
      }

      if (this.sp.length > 0 && this.sp[0] === this.root) {
        throw "the head of the sp cannot be the root";
      }

      for (var i = this.sp.length - 1; i > 1; i++) {
        if (this.sp[i].parent !== this.sp[i - 1]) {
          throw "wrong parent for a node in the sp";
        }
      }

      for (var i = 0; i < this.nonfullnodes.length; i++) {
        var node = this.nonfullnodes[i];
        var nodeSize = node.childNodes.length;
        if (nodeSize === this.maxChildren) {
          throw "nonfull internal node is full";
        }

        if (node !== this.root && nodeSize < this.minChildren) {
          throw "nonfull internal node has too few elements";
        }
      }

      this._validate(this.root);
    };

    /**
     * Implementation of the validate method.
     * @param {BushInternalNode|BushExternalNode} node node to validate.
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._validate = function(node) {
      if (!this._isLeaf(node)) {
        if (node.childNodes.length > this.maxChildren) {
          throw "node has too many children";
        }

        for (var i = 0; i < node.childNodes.length; i++) {
          if (node.childNodes[i].parent !== node) {
            throw "node has wrong parent";
          }
        }

        for (var i = 0; i < node.childNodes.length; i++) {
          this._validate(node.childNodes[i]);
        }
      }
    };

    /**
     * Return a string representation of the tree.
     * @returns {String} the string representation of the tree.
     * @memberOf Snake.Crypto.Bush#
     */
    Bush.prototype.printTree = function() {
      return this._printTree(this.root, "", true);
    };

    /**
     * Implementation of the printTree method.
     * @param {BushLeafNode|BushInternalNode} node
     * @param {String} prefix
     * @param {Boolean} isTail
     * @memberOf Snake.Crypto.Bush#
     * @private
     */
    Bush.prototype._printTree = function(node, prefix, isTail) {
      var str = prefix.concat((isTail ? "└── " : "├── "));

      if (!this._isLeaf(node)) {
        str = str.concat("k" + node.nodeKey.id + "\n");

        for (var i = 0; i < node.childNodes.length - 1; i++) {
          var childStr = this._printTree(node.childNodes[i], prefix +
              (isTail ? "    " : "│   "), false);
          str = str.concat(childStr);
        }
        if (node.childNodes.length >= 1) {
          var lastNode = node.childNodes[node.childNodes.length - 1];
          var childStr = this._printTree(lastNode, prefix + (isTail ?
              "    " : "│   "), true);
          str = str.concat(childStr);
        }
      } else {
        str = str.concat("u" + node.nodeKey.id + " - " + node.nodeId + "\n");
      }

      return str;
    };

    /**
     * Serializes the tree.
     * @returns {String} the serialized object.
     */
    Bush.prototype.stringify = function() {
      // removes arrays with reference to nodes
      delete this.sp;
      delete this.nonfullnodes;
      delete this.externalNodes;

      // traverses the tree and removes all the parent attribute
      var parentRemover = function(node) {
        delete node.parent;

        if (node.childNodes) {
          for (var i = 0; i < node.childNodes.length; i++) {
            parentRemover(node.childNodes[i]);
          }
        }
      };
      parentRemover(this.root);

      return this;
    };

    /**
     * Deserializes a Bush tree.
     * @param {String} tree the serialized tree.
     */
    Bush.prototype.parse = function(tree) {
      for (var key in tree) {
        this[key] = tree[key];
      }

      this.sp = [];
      this.nonfullnodes = [];
      this.externalNodes = {};

      var bushRestorer = function(node) {
        if (node) {
          if (node.childNodes && node.nodeKey) {
            // the node is internal
            node.parent = null;

            for (var i = 0; i < node.childNodes.length; i++) {
              // sets the parent reference in the child nodes
              node.childNodes[i].parent = node;
            }

            // if the node is too empty insert it into the special path
            if (node.childNodes.length < this.minChildren) {
              this.sp.push(node);
            }

            // if the node is not full insert it into the nonfullnodes
            if (node.childNodes.length >= this.minChildren &&
                node.childNodes.length < this.maxChildren) {
              this.nonfullnodes.push(node);
            }

            for (var i = 0; i < node.childNodes.length; i++) {
              bushRestorer(node.childNodes[i]);
            }
          } else if (node.nodeId && node.nodeKey) {
            this.externalNodes[node.nodeId] = node;
          }
        }
      }.bind(this);

      bushRestorer(this.root);
    };

    return Bush;
  }());
}());
