(function() {
  "use strict";

  // http://jsperf.com/new-date-vs-date-now-vs-performance-now/11
  // Using a cached window.performance object seems to be the fastest method to
  // call window.performance.now. Date.now is not used because its time
  // resolution is limited to 1 millisecond.
  var perf = window.performance;

  /**
   * Helper function to force serialized execution. We use this on functions
   * using the WebCrypto API which first import the key and then use it, since,
   * if this process is not serialized (e.g. importKey1, importKey2, sign1,
   * sign2 instead of importKey1, sign1, importKey2, sign2) we get weird errors
   * under Firefox.
   *
   * @param func {Function} the function that must be serialized.
   *
   * @return {Function} a function whose execution is serialized.
   */
  function locked(func) {
    var lock = Snake.PromiseQueue();

    return function () {
      var acquiredLock = lock.enqueue();
      var args = arguments;

      return acquiredLock.wait
          .then(function () { return func.apply(this, args); }.bind(this))
          .catch(function (e) { acquiredLock.unlock(); throw e; })
          .then(function (res) { acquiredLock.unlock(); return res; })
    };
  }

  var cryptoConfig;
  var cryptoCounter;

  var cryptoImpl;

  function abvToUint8Array(abv) {
    return new Uint8Array(abv);
  }

  function exportKeyPair(e) {
    return Promise.every(
        cryptoImpl.exportKey("jwk", e.publicKey),
        cryptoImpl.exportKey("jwk", e.privateKey)
        ).then(function(results) {
      return {
        publicKey: results[0],
        privateKey: results[1]
      };
    });
  }

  function isWebCryptoKey(key) {
    return key instanceof CryptoKey;
  }

  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    init: function() {
      cryptoConfig = Snake.Config.Crypto;
      cryptoCounter = Snake.Perf.CryptoCounter;

      cryptoImpl = window.crypto.subtle;
    },
    Common: {
      compareHashes: function(a, b) {
        return util.memcmp(a, b);
      },
      /**
       * Helper function to normalize ArrayBufferView, Base128 and WebCryptoAPI
       * keys to WebCryptoAPI keys
       * @param {ArrayBufferView|Base128|Key} key key to import, it can be:
       *        * Key: just forwarded
       *        * String (Base128): decoded & imported
       *        * ArrayBufferView: imported
       * @param {String} type type of the key.
       * @param {String} algorithm destination algorithm.
       * @param {Array<String>} usages possible usages for the imported key.
       * @returns {Promise<Key>} the WebCryptoAPI imported key.
       */
      importKey: function(key, type, algorithm, usages) {
        // if key is a WebCrypto key just return it
        if (isWebCryptoKey(key)) {
          return Snake.identityPromise(key);
        } else {
          return cryptoImpl.importKey(type, key, algorithm, true, usages);
        }
      },
      /**
       * Convert a raw elliptic curve public key in (uncompressed) ANSI format
       * to the JWK format.
       * @param {ArrayBufferView} key a key in the uncompressed ANSI format.
       * @returns {Object} a JWK public key object.
       */
      ecAnsiPublicKeyToJwk: function(key) {
        if (!key[0] === 0x04) {
          throw new Error("Only uncompressed format is supported");
        }

        key = key.subarray(1);
        var length = key.length / 2;

        return {
          kty: "EC",
          crv: cryptoConfig.curve,
          x: util.b64encode(key.subarray(0, length)),
          y: util.b64encode(key.subarray(length, length * 2))
        };
      },
      jwkToAnsiPublicKey: function(key) {
        if (key.kty !== "EC" || key.crv !== cryptoConfig.curve) {
          throw new Error("Unexpected curve");
        }

        return Snake.concatBuffers(new Uint8Array([4]), util.b64decode(key.x),
            util.b64decode(key.y));
      },
      privateKeyToJwk: function(privateKey, publicKey) {
        var length;
        var result = {
          kty: "EC",
          crv: cryptoConfig.curve,
          d: util.b64encode(privateKey)
        };

        // The public key part is optional
        if (!_.isUndefined(publicKey)) {
          if (!publicKey[0] === 0x04) {
            throw new Error("Only uncompressed format is supported");
          }

          publicKey = publicKey.subarray(1);
          length = publicKey.length / 2;

          result["x"] = util.b64encode(publicKey.subarray(0, length));
          result["y"] = util.b64encode(publicKey.subarray(length, length * 2));
        }

        return result;
      },
      jwkToPrivateKey: function(key) {
        if (key.kty !== "EC" || key.crv !== cryptoConfig.curve) {
          throw new Error("Unexpected curve");
        }

        return util.b64decode(key.d);
      },
      /**
       * Hash as string.
       * @memberOf Snake.Crypto.Common
       * @param {ArrayBufferView} data to hash.
       * @returns {Promise<ArrayBufferView>} the hash.
       */
      hash: function(data) {
        var start = perf.now();

        return cryptoImpl.digest(cryptoConfig.hashAlgorithm, data)
            .then(abvToUint8Array)
            .then(function(e) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.hashAlgorithm,
              data.length, start, end);
          return e;
        }).catch(Snake.errorHandler.bind("crypto"));
      },
      /**
       * Generates an HMAC key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the exported key.
       */
      generateHMACKey: function() {
        return cryptoImpl.generateKey({
          "name": cryptoConfig.hmacAlgorithm,
          "hash": cryptoConfig.hashAlgorithm,
          "length": cryptoConfig.hmacKeyLength
        },
        true,
        ["sign", "verify"]).then(function(e) {
          return cryptoImpl.exportKey("raw", e);
        }).then(abvToUint8Array)
        .catch(Snake.errorHandler.bind("crypto"));
      },
      /**
       * Computes the hash of secret and uses is as key to compute the MAC
       * of data.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} secret key to be hashed.
       * @param {ArrayBufferView} data data on which compute the MAC.
       * @returns {Promise<ArrayBufferView>} the MAC of data using the hash of
       *          secret as key.
       */
      macWithHashedKey: function(secret, data) {
        var start = perf.now();

        // Compute plain hash of secret
        return cryptoImpl.digest(cryptoConfig.hashAlgorithm, secret)
            .then(abvToUint8Array)
            .then(function(secretHash) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.hashAlgorithm,
              secret.length, start, end);
          return this.mac(secretHash, data);
        }.bind(this)).catch(Snake.errorHandler.bind("crypto"));
      },
      /**
       * Computes the MAC of data using key as key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key.
       * @param {ArrayBufferView} data the data on which compute the MAC.
       * @returns {Promise<ArrayBufferView>} the MAC of the data.
       */
      mac: locked(function(key, data) {
        var start = perf.now();

        // Import the hashed secret as a key
        return this.importKey(key, "raw", {
          "name": cryptoConfig.hmacAlgorithm,
          "hash": cryptoConfig.hashAlgorithm
        }, ["sign"]).then(function(importedKey) {
          // Sign the data to authenticate with the hash of the secret
          return cryptoImpl.sign({
            "name": cryptoConfig.hmacAlgorithm,
            "hash": cryptoConfig.hashAlgorithm
          }, importedKey, data);
        }).then(abvToUint8Array)
        .then(function(e) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.hmacAlgorithm,
              data.length, start, end);
          return e;
        }).catch(Snake.errorHandler.bind("crypto"));
      }),
      /**
       * Generates a symmetric key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the exported key.
       */
      generateSymmetricKey: function() {
        return cryptoImpl.generateKey({
          "name": cryptoConfig.symmAlgo,
          "length": cryptoConfig.symmKeyLength
        },
        true, /* extractable */
        ["encrypt", "decrypt"]).then(function(e) {
          return cryptoImpl.exportKey("raw", e);
        }).then(abvToUint8Array)
        .catch(Snake.errorHandler.bind("crypto"));
      },
      /**
       * Encrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key to be used to encrypt
       *                                      the data.
       * @param {ArrayBufferView} data the data to be encrypted.
       * @param {ArrayBufferView} iv the initialization vector.
       * @returns {Promise<ArrayBufferView>} the encrypted data.
       */
      symmetricEncryption: locked(function(key, data, iv) {
        var start = perf.now();

        return this.importKey(key, "raw", cryptoConfig.symmAlgo, ["encrypt"])
            .then(function(importedKey) {
          return cryptoImpl.encrypt({
            "name": cryptoConfig.symmAlgo,
            "iv": iv
          }, importedKey, data);
        }).then(abvToUint8Array)
        .then(function(e) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.symmAlgo + "-encrypt",
              data.lenght, start, end);
          return e;
        }).catch(Snake.errorHandler.bind("crypto"));
      }),
      /**
       * Decrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key used to decrypt the
       *                                          data.
       * @param {ArrayBufferView} data data to decrypt.
       * @param {ArrayBufferView} iv initialization vector.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmetricDecryption: locked(function(key, data, iv) {
        var start = perf.now();

        return this.importKey(key, "raw", cryptoConfig.symmAlgo, ["decrypt"])
            .then(function(importedKey) {
          return cryptoImpl.decrypt({
            "name": cryptoConfig.symmAlgo,
            "iv": iv
          }, importedKey, data);
        }).then(abvToUint8Array)
        .then(function(e) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.symmAlgo + "-decrypt",
              data.length, start, end);
          return e;
        }).catch(Snake.errorHandler.bind("crypto"));
      }),
      /**
       * Generates a symmetric key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the eported key.
       */
      generateSymmAuthKey: function() {
        return cryptoImpl.generateKey({
          "name": cryptoConfig.symmAuthAlgo,
          "length": cryptoConfig.symmAuthKeyLength
        },
        true,
        ["encrypt", "decrypt"]).then(function(e) {
          return cryptoImpl.exportKey("raw", e);
        }).then(abvToUint8Array)
        .catch(Snake.errorHandler.bind("crypto"));
      },
      /**
       * Encrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the exported key to use.
       * @param {ArrayBufferView} data the data to be encrypted.
       * @param {ArrayBufferView} iv initialization vector.
       * @param {ArrayBufferView} additionalData data that will not end in the
       *        ciphertext but will be authenticated.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmAuthEncryption: locked(function(key, data, iv, additionalData) {
        var start = perf.now();

        // XXX netflix plugin is not able to handle empty typed array
        additionalData = additionalData || new Uint8Array(1);

        return this.importKey(key, "raw", cryptoConfig.symmAuthAlgo, ["encrypt"])
            .then(function(importedKey) {
          return cryptoImpl.encrypt({
            "name": cryptoConfig.symmAuthAlgo,
            "iv": iv,
            "additionalData": additionalData,
            "tagLength": cryptoConfig.symmAuthTagLength
          }, importedKey, data);
        }).then(abvToUint8Array)
        .then(function(encryptedData) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.symmAuthAlgo + "-encrypt",
              data.length + additionalData.length, start, end);
          return encryptedData;
        });
      }),
      /**
       * Decrypts an object.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the exported key to be used to
       *        encrypt the data.
       * @param {ArrayBufferView} data the string containing the encrypted data.
       * @param {ArrayBufferView} iv initialization vector.
       * @param {ArrayBufferView} additionalData data not part of the
       *        ciphertext but to include for the MAC verification.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmAuthDecryption: locked(function(key, data, iv, additionalData) {
        var start = perf.now();

        // XXX netflix plugin is not able to handle empty typed array
        additionalData = additionalData || new Uint8Array(1);

        return this.importKey(key, "raw", cryptoConfig.symmAuthAlgo, ["decrypt"])
            .then(function(importedKey) {
          return cryptoImpl.decrypt({
            "name": cryptoConfig.symmAuthAlgo,
            "iv": iv,
            "additionalData": additionalData,
            "tagLength": cryptoConfig.symmAuthTagLength
          }, importedKey, data);
        }).then(abvToUint8Array)
        .then(function(decryptedData) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.symmAuthAlgo + "-decrypt",
              data.length + additionalData.length, start, end);
          return decryptedData;
        });
      }),
      /**
       *
       * @param {ArrayBufferView} baseJwkKey
       * @param {ArrayBufferView} otherPublicPoint
       * @param {Array<String>} usages
       * @returns {Promise<Key>} the derived key.
       */
      deriveKey: locked(function(baseJwkKey, otherPublicPoint, usages) {
        var baseKey;
        var start = perf.now();

        return this.importKey(baseJwkKey,
                              "jwk",
                              cryptoConfig.keyAgreementAlgorithm,
                              ["deriveKey"])
            .then(function(importedBaseKey) {
          baseKey = importedBaseKey;
          return this.importKey(otherPublicPoint,
                                "jwk",
                                cryptoConfig.keyAgreementAlgorithm,
                                []);
        }.bind(this)).then(function(otherPublicKey) {
          return cryptoImpl.deriveKey({
            "name": cryptoConfig.keyAgreementAlgorithm.name,
            "public": otherPublicKey
          },
          baseKey, {
            "name": cryptoConfig.symmAuthAlgo,
            "length": cryptoConfig.symmAuthKeyLength
          },
          false,
          usages);
        }).then(function(derivedKey) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.keyAgreementAlgorithm + "derive",
              0, start, end);
          return derivedKey;
        }).catch(Snake.errorHandler.bind("crypto"));
      }),
      /**
       * Password-based key derivation using PBKDF2 algorithm.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} password the password.
       * @param {ArrayBufferView} salt the salt used to derive the key.
       * @returns {Promise<ArrayBufferView>} the derived key.
       */
      deriveKeyFromPassword: locked(function(password, salt) {
        var start = perf.now();

        return cryptoImpl.importKey("raw",
                                    password,
                                    {"name": 'PBKDF2'},
                                    false,
                                    ["deriveKey"]).then(function (key) {

          return cryptoImpl.deriveKey({
              "name": "PBKDF2",
              "salt": salt,
              "iterations": cryptoConfig.passDerivIters,
              "hash": cryptoConfig.passDerivHash
            },
            key,
            {
              "name": cryptoConfig.symmAuthAlgo,
              "length": cryptoConfig.symmAuthKeyLength
            },
            false,
            ["encrypt", "decrypt"]);
        })
        .then(function(derivedKey) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.passDerivAlgo, 0, start, end);
          return derivedKey;
        });
      }),
      /**
       * Sign data with the specified key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} data the data to sign.
       * @param {ArrayBufferView|Base128|Key} key private key to be used to sign
       *        the data;
       * @returns {Promise<ArrayBufferView>} the signed data.
       */
      sign: locked(function(data, key) {
        var start = perf.now();

        return Snake.identityPromise().then(function () {
          return this.importKey(key,
                                "jwk",
                                cryptoConfig.signAlgorithm,
                                ["sign"]);
        }.bind(this))
        .then(function(importedKey) {
          return cryptoImpl.sign({
            "name": cryptoConfig.signAlgorithm.name,
            "hash": cryptoConfig.hashAlgorithm
          }, importedKey, data)
          .then(abvToUint8Array)
          .then(function(e) {
            var end = perf.now();
            cryptoCounter.stat(cryptoConfig.signAlgorithm + "-sign",
                data.length, start, end);
            return e;
          });
        })
        .catch(Snake.errorHandler.bind("crypto"));
      }),
      /**
       * Verify data with the specified key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} data the data to verify.
       * @param {ArrayBufferView} signature the signature to check against.
       * @param {ArrayBufferView|Base128|Key} key public key to be used to
       *        verify the data;
       */
      verify: locked(function(data, signature, key) {
        var start = perf.now();

        return this.importKey(key,
                              "jwk",
                              cryptoConfig.signAlgorithm,
                              ["verify"])
            .then(function(importedKey) {
          return cryptoImpl.verify({
            "name": cryptoConfig.signAlgorithm.name,
            "hash": cryptoConfig.hashAlgorithm
          }, importedKey, signature, data).then(function(e) {
            var end = perf.now();
            cryptoCounter.stat(cryptoConfig.signAlgorithm + "-verify",
                data.length, start, end);
            return e;
          });
        }).catch(Snake.errorHandler.bind("crypto"));
      }),
      generateIV: function() {
        // TODO: use generateKey
        // TODO: AES-GCM allows up to ~2^56 bytes of IV, does it make sense
        // to use more than a block size?
        var start = perf.now();
        var iv = window.crypto.getRandomValues(
            new Uint8Array(cryptoConfig.IVlength));
        var end = perf.now();

        cryptoCounter.stat("IV", cryptoConfig.IVlength, start, end);

        return Snake.identityPromise(iv);
      },
      generatePasswordSalt: function() {
        var start = perf.now();
        var passwordSalt = window.crypto.getRandomValues(
            new Uint8Array(cryptoConfig.passwordSaltLength));
        var end = perf.now();

        cryptoCounter.stat("SALT", cryptoConfig.passwordSaltLength, start, end);

        return Snake.identityPromise(passwordSalt);
      },
      generateSigningKeyPair: function() {
        var start = perf.now();

        return cryptoImpl.generateKey(
          cryptoConfig.signAlgorithm,
          true, /* extractable */
          ["sign", "verify"]
        ).then(exportKeyPair).then(function(result) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.signAlgorithm + "-generate",
                0, start, end);

          return result;
        }).catch(Snake.errorHandler.bind("crypto"));
      },
      generateKeyAgreementKeyPair: function() {
        var start = perf.now();

        return cryptoImpl.generateKey(cryptoConfig.keyAgreementAlgorithm,
                                      true, /* extractable */
                                      ["deriveKey"])
          .then(exportKeyPair).then(function(result) {
          var end = perf.now();
          cryptoCounter.stat(cryptoConfig.keyAgreementAlgorithm + "-generate",
                0, start, end);

          return result;
        }).catch(Snake.errorHandler.bind("crypto"));
      }
    }
  });
}());
