(function() {
  "use strict";

  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    FakeCommon: {
      prng: 0,
      init: function() {
        // Replace Snake.Crypto.Common with myself
        Snake.Crypto.Common = Snake.Crypto.FakeCommon;

        // Set fake friendship protocol handler
        Snake.Crypto.FriendshipProtocol.handleProtocols = this.handleProtocols;

        // Set fake initial message for friendship protcol
        Snake.Crypto.SMP.step1 = function (otherId, isAuthenticated) {
          return Snake.identityPromise().then(function () {
            if (!isAuthenticated) {
              return Snake.Crypto.SMP.askQuestion(otherId.toString());
            }
          }).then(function () {
            return {"send": null, "store": null};
          });
        };
        Snake.Crypto.FHMQV.initiate = function() {
          var identity = new Uint8Array(32);
          identity[0] = Snake.sessionUser.id;
          return Snake.identityPromise({
            "send": {
              ephemeralPublicKey: new Uint8Array([4]),
              confirmationToken: new Uint8Array(32),
              identity: identity,
              type: Snake.Crypto.FHMQV.MessageType.InitialRequest
            },
            "store": {
              common: {
                key: new Uint8Array(32),
                signingKeyHash: new Uint8Array(32)
              }
            }
          });
        };

        // Set fake FHMQV initial request handler
        Snake.Crypto.FHMQV.handleInitialRequest = function(receivedData) {
          _.extend(receivedData.fhmqv, {
            _confirmationToken: new Uint8Array(32),
            _key: new Uint8Array(32),
            _signingKeyHash: new Uint8Array(32)
          });
          receivedData._senderId = receivedData.fhmqv.identity[0];
          receivedData._recipientType = "user";
          receivedData._recipientId = Snake.sessionUser.id;

          return Snake.identityPromise();
        };
      },
      nextRandom: function() {
        this.prng += 1;
        return this.prng;
      },
      toBytes: function(number, array) {
        var i = 0;
        while (number != 0 && i < array.length) {
          array[i] = number & 0xff;
          number = number >>> 8;
          i += 1;
        }

        return array;
      },
      compareHashes: function(a, b) {
        return true;
      },
      handleProtocols: function(handlers, message, friendship, senderId) {
        // TODO: can't use this.messageHandler because, for some reason, we're
        //       binded to Snake.Crypto.FriendshipProtocol
        return _(handlers).map(Snake.Crypto.FakeCommon.messageHandler.bind(null, message));
      },
      messageHandler: function(message, handler) {
        var toSend = null;
        if (handler.fieldName === "fhmqv"
           &&  message.get("fhmqv").type === 0) {
          toSend = {
            confirmationToken: new Uint8Array(32),
            type: Snake.Crypto.FHMQV.MessageType.ConfirmationToken
          };
        }

        return Snake.identityPromise({
          "store": {
            common: {
              key: new Uint8Array(32),
              signingKeyHash: new Uint8Array(32)
            }
          },
          "send": toSend,
          "done": true
        });
      },
      /**
       * Helper function to normalize ArrayBufferView, Base128 and WebCryptoAPI
       * keys to WebCryptoAPI keys
       * @param {ArrayBufferView|Base128|Key} key key to import, it can be:
       *        * Key: just forwarded
       *        * String (Base128): decoded & imported
       *        * ArrayBufferView: imported
       * @param {String} type type of the key.
       * @param {String} algorithm destination algorithm.
       * @returns {Promise<Key>} the WebCryptoAPI imported key.
       */
      importKey: function(key, type, algorithm) {
        var fakeImportedKey = {key: "fake"};
        return Snake.identityPromise(fakeImportedKey);
      },
      /**
       * Convert a raw elliptic curve public key in (uncompressed) ANSI format
       * to the JWK format.
       * @param {ArrayBufferView} key a key in the uncompressed ANSI format.
       * @returns {Object} a JWK public key object.
       */
      ecAnsiPublicKeyToJwk: function(key) {
        if (!key[0] === 0x04) {
          throw new Error("Only uncompressed format is supported");
        }

        key = key.subarray(1);
        var length = key.length / 2;

        return util.utf8encode(JSON.stringify({
          kty: "EC",
          crv: "P-256",
          x: util.b64encode(key.subarray(0, length)),
          y: util.b64encode(key.subarray(length, length * 2))
        }));
      },
      jwkToAnsiPublicKey: function(key) {
        key = JSON.parse(util.utf8decode(key));
        if (key.kty !== "EC" || key.crv !== "P-256") {
          throw new Error("Unexpected curve");
        }

        return Snake.concatBuffers(new Uint8Array([4]), util.b64decode(key.x),
            util.b64decode(key.y));
      },
      privateKeyToJwk: function(privateKey, publicKey) {
        if (!publicKey[0] === 0x04) {
          throw new Error("Only uncompressed format is supported");
        }

        publicKey = publicKey.subarray(1);
        var length = publicKey.length / 2;

        return util.utf8encode(JSON.stringify({
          kty: "EC",
          crv: "P-256",
          d: util.b64encode(privateKey),
          x: util.b64encode(publicKey.subarray(0, length)),
          y: util.b64encode(publicKey.subarray(length, length * 2))
        }));
      },
      jwkToPrivateKey: function(key) {
        key = JSON.parse(util.utf8decode(key));
        if (key.kty !== "EC" || key.crv !== "P-256") {
          throw new Error("Unexpected curve");
        }

        return util.b64decode(key.d);
      },
      /**
       * Hash as string.
       * @memberOf Snake.Crypto.Common
       * @param {ArrayBufferView} data to hash.
       * @returns {Promise<ArrayBufferView>} the hash.
       */
      hash: function(data) {
        var i;
        var fakeHash = new Uint8Array(32);
        for (i = 0; i < data.length; i++)
          fakeHash[i % 32] += data[i];
        return Snake.identityPromise(fakeHash);
      },
      /**
       * Generates an HMAC key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the exported key.
       */
      generateHMACKey: function() {
        var fakeKey = this.toBytes(this.nextRandom(), new Uint8Array(32));
        return Snake.identityPromise(fakeKey);
      },
      /**
       * Computes the hash of secret and uses is as key to compute the MAC
       * of data.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} secret key to be hashed.
       * @param {ArrayBufferView} data data on which compute the MAC.
       * @returns {Promise<ArrayBufferView>} the MAC of data using the hash of
       *          secret as key.
       */
      macWithHashedKey: function(secret, data) {
        var i;
        var fakeHash = new Uint8Array(32);
        for (i = 0; i < data.length; i++)
          fakeHash[i % 32] += data[i];
        for (i = 0; i < secret.length; i++)
          fakeHash[31 - i % 32] += secret[i];
        return Snake.identityPromise(fakeHash);
      },
      /**
       * Computes the MAC of data using key as key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key.
       * @param {ArrayBufferView} data the data on which compute the MAC.
       * @returns {Promise<ArrayBufferView>} the MAC of the data.
       */
      mac: function(key, data) {
        var i = 0;
        var fakeMac = new Uint8Array(32);
        for (i = 0; i < data.length; i++)
          fakeMac[i % 32] += data[i];
        for (i = 0; i < key.length; i++)
          fakeMac[31 - i % 32] += key[i];
        return Snake.identityPromise(fakeMac);
      },
      /**
       * Generates a symmetric key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the exported key.
       */
      generateSymmetricKey: function() {
        var fakeKey = this.toBytes(this.nextRandom(), new Uint8Array(32));
        return Snake.identityPromise(fakeKey);
      },
      /**
       * Encrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key to be used to encrypt
       *                                      the data.
       * @param {ArrayBufferView} data the data to be encrypted.
       * @param {ArrayBufferView} iv the initialization vector.
       * @returns {Promise<ArrayBufferView>} the encrypted data.
       */
      symmetricEncryption: function(key, data, iv) {
        return Snake.identityPromise(data);
      },
      /**
       * Decrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the key used to decrypt the data.
       * @param {ArrayBufferView} data data to decrypt.
       * @param {ArrayBufferView} iv initialization vector.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmetricDecryption: function(key, data, iv) {
        return Snake.identityPromise(data);
      },
      /**
       * Generates a symmetric key.
       * @memberOf Snake.Crypto.Common#
       * @returns {Promise<ArrayBufferView>} the eported key.
       */
      generateSymmAuthKey: function() {
        var fakeKey = this.toBytes(this.nextRandom(), new Uint8Array(32));
        return Snake.identityPromise(fakeKey);
      },
      /**
       * Encrypts a string.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the exported key to use.
       * @param {ArrayBufferView} data the data to be encrypted.
       * @param {ArrayBufferView} iv initialization vector.
       * @param {ArrayBufferView} additionalData data that will not end in the
       *        ciphertext but will be authenticated.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmAuthEncryption: function(key, data, iv, additionalData) {
        return Snake.identityPromise(data);
      },
      /**
       * Decrypts an object.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView|Base128|Key} key the exported key to be used to
       *        encrypt the data.
       * @param {ArrayBufferView} data the string containing the encrypted data.
       * @param {ArrayBufferView} iv initialization vector.
       * @param {ArrayBufferView} additionalData data not part of the
       *        ciphertext but to include for the MAC verification.
       * @returns {Promise<ArrayBufferView>} the decrypted data.
       */
      symmAuthDecryption: function(key, data, iv, additionalData) {
        return Snake.identityPromise(data);
      },
      /**
       *
       * @param {ArrayBufferView} baseJwkKey
       * @param {ArrayBufferView} otherPublicPoint
       * @returns {Promise<Key>} the derived key.
       */
      deriveKey: function(baseJwkKey, otherPublicPoint) {
        var fakeKey = new Uint8Array(32);
        return Snake.identityPromise(fakeKey);
      },
      /**
       * Password-based key derivation using PBKDF2 algorithm.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} password the password.
       * @param {ArrayBufferView} salt the salt used to derive the key.
       * @returns {Promise<ArrayBufferView>} the derived key.
       */
      deriveKeyFromPassword: function(password, salt) {
        var fakeKey = new Uint8Array(32);
        return Snake.identityPromise(fakeKey);
      },
      /**
       * Sign data with the specified key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} data the data to sign.
       * @param {ArrayBufferView|Base128|Key} key private key to be used to sign
       *        the data;
       * @returns {Promise<ArrayBufferView>} the signed data.
       */
      sign: function(data, key) {
        var i;
        var fakeSign = new Uint8Array(32);
        for (i = 0; i < data.length; i++)
          fakeSign[i % 32] += data[i];
        for (i = 0; i < key.length; i++)
          fakeSign[i % 32] += key[i];
        return Snake.identityPromise(fakeSign);
      },
      /**
       * Verify data with the specified key.
       * @memberOf Snake.Crypto.Common#
       * @param {ArrayBufferView} data the data to verify.
       * @param {ArrayBufferView} signature the signature to check against.
       * @param {ArrayBufferView|Base128|Key} key public key to be used to
       *        verify the data;
       */
      verify: function(data, signature, key) {
        return Snake.identityPromise(true);
      },
      generateIV: function() {
        var iv = window.crypto.getRandomValues(new Uint8Array(16));
        return Snake.identityPromise(iv);
      },
      generatePasswordSalt: function() {
        var passwordSalt = window.crypto.getRandomValues(new Uint8Array(32));
        return Snake.identityPromise(passwordSalt);
      },
      generateSigningKeyPair: function() {
        var fakePrivateKey = new Uint8Array(32);
        var fakePublicKey = new Uint8Array(65);
        fakePublicKey[0] = 4;

        var fakeJwkPublicKey = this.ecAnsiPublicKeyToJwk(fakePublicKey);
        var fakeJwkPrivateKey = this.privateKeyToJwk(fakePrivateKey,
            fakePublicKey);

        var fakeJwkKey = {
          publicKey: fakeJwkPublicKey,
          privateKey: fakeJwkPrivateKey
        };

        return Snake.identityPromise(fakeJwkKey);
      },
      generateKeyAgreementKeyPair: function() {
        var fakePrivateKey = new Uint8Array(32);
        var fakePublicKey = new Uint8Array(65);
        fakePublicKey[0] = 4;

        var fakeJwkPublicKey = this.ecAnsiPublicKeyToJwk(fakePublicKey);
        var fakeJwkPrivateKey = this.privateKeyToJwk(fakePrivateKey,
            fakePublicKey);

        var fakeJwkKey = {
          publicKey: fakeJwkPublicKey,
          privateKey: fakeJwkPrivateKey
        };

        return Snake.identityPromise(fakeJwkKey);
      }
    }
  });
}());
