(function() {
  "use strict";
  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    /**
     * Protocol for authenticated and confirmed key derivation based on FHMQV-C.
     * @class Snake.Crypto.FHMQV
     */
    // TODO: find a more suitable format for keys instead than hex (base64?)
    // TODO: create a WebCryptoAPI-like API for FHMQV and SMP
    FHMQV: _(_(Snake.Crypto.FriendshipProtocol).clone()).extend({
      MessageType: {
        InitialRequest: 0,
        ConfirmationToken: 1
      },
      MessageHandler: [
        "answer",
        "checkConfirmationToken"
      ],
      /**
       * Expands the input message with the fields for the other end of the
       * handshake and returns the information which have to be stored in the
       * friendship descriptor for future use.
       *
       * @memberOf Snake.Crypto.FHMQV
       * @param {Number} otherId ID of the other end
       * @returns {Promise} a Promise that will be resolved when all the
       *          computations are done. The result of the promise is an hash
       *          containing a `store` field with the data to be stored in
       *          the current user's private descriptor and a field `send`
       *          containing the data to send to the other end.
       *
       * The `store` field contains:
       *
       * * `key`: the derived symmetric key
       * * "expectedConfirmationToken": the token which the other end
       *   should send in the answer.
       *
       * The `send` field contains:
       *
       *  * `ephemeralPublicKey`: the public part of the used
       *    ephemeral key
       *  * `confirmationToken`: the current user's confirmation token
       *  * `identity`: the identity of the current user encrypted
       *    with K0
       */
      initiate: function(otherId) {
        var other, myEphemeralKey, myEphemeralKeyJwk,
            otherEphemeralPublicKey, otherStaticPublicKey, myStaticKey,
            myUserIdHex, signingKeyHashPromise, encryptedIdentity, fhmqv, k2,
            myConfirmationToken, otherConfirmationToken;

        other = new Snake.Model.User({id: otherId}, {});

        // Fetch other user's data
        return other.fetch().result.then(function() {
          var curve;
          var otherEphemeralPublicKeyJwk;

          // Compute the hash of the signing key for storage
          // We'll wait for this promise later
          signingKeyHashPromise = Snake.Crypto.Common.hash(
              other.get("signingPublicKey"));

          // TODO: do we like hex?
          // Convert user's ID to hex, used several times
          myUserIdHex = Snake.sessionUser.id.toString(16);

          // Get the chosen curve
          curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;

          // Load other user's public keys
          otherStaticPublicKey = curve.decodePointHex(
              util.abv2hex(other.get("staticPublicKey")));
          otherEphemeralPublicKey = curve.decodePointHex(
              util.abv2hex(other.get("ephemeralPublicKey")));

          // Validate the keys before continuing
          // TODO: handle error
          otherStaticPublicKey.validate();
          otherEphemeralPublicKey.validate();

          // Generate a new ephemeral key pair
          myEphemeralKey = new ECKey(Snake.Config.Crypto.curve);
          myEphemeralKey.generate();

          // Load the static key
          myStaticKey = new ECKey(Snake.Config.Crypto.curve,
              util.abv2hex(Snake.sessionUser.get("staticPrivateKey")),
              util.abv2hex(Snake.sessionUser.get("staticPublicKey")));

          // XXX
          myEphemeralKeyJwk = {
            kty: "EC",
            crv: Snake.Config.Crypto.curve,
            d: util.b64encode(util.hex2abv(myEphemeralKey.getPriv().toHex())),
            x: util.b64encode(util.hex2abv(myEphemeralKey.getPubPoint().getX().toHex())),
            y: util.b64encode(util.hex2abv(myEphemeralKey.getPubPoint().getY().toHex()))
          };

          otherEphemeralPublicKeyJwk =
              Snake.Crypto.Common.ecAnsiPublicKeyToJwk(otherEphemeralPublicKey.serializeANSI());

          // Import the ephemeral private key into the WebCryptoAPI and derive K0.
          return Snake.Crypto.Common.deriveKey(myEphemeralKeyJwk,
                                               otherEphemeralPublicKeyJwk,
                                               ["encrypt"]);
        }.bind(this)).then(function(k0) {
          // TODO: put a real IV
          // TODO: use AES-CBC not AES-GCM
          // Encrypt current user's ID using K0
          return Snake.Crypto.Common.symmAuthEncryption(k0,
              util.hex2abv(myUserIdHex), new Uint8Array(32));
        }.bind(this)).then(function(result) {
          encryptedIdentity = result;

          // Derive the authenticated session key K2 and current
          // user's confirmation token
          fhmqv = new FHMQV(Snake.Config.Crypto.curve,
              CryptoJS.algo[Snake.Config.Crypto.hashAlgorithm.replace("-", "")],
              CryptoJS.algo[Snake.Config.Crypto.hmacAlgorithm],
              true /* initiator */,
              myEphemeralKey.getPriv(), myEphemeralKey.getPubPoint(),
              otherEphemeralPublicKey,
              myStaticKey.getPriv(), myStaticKey.getPubPoint(),
              otherStaticPublicKey,
              Snake.sessionUser.getDataToAuthenticate(),
              other.getDataToAuthenticate(),
              false /* validate */);

          myConfirmationToken = fhmqv.getMyConfirmationToken();
          otherConfirmationToken = fhmqv.getOtherConfirmationToken();
          k2 = fhmqv.getK2();

          // Return signingKeyHashPromise to get the result of the previously
          // launched hash computation
          return signingKeyHashPromise;
        }.bind(this)).then(function(signingKeyHash) {
          // Return the fields to update in the friendship descriptor
          // when the message has been successufully sent
          return {
            send: {
              ephemeralPublicKey: util.hex2abv(
                  myEphemeralKey.getPubPoint().encodeHex()),
              confirmationToken: myConfirmationToken,
              identity: encryptedIdentity,
              type: Snake.Crypto.FHMQV.MessageType.InitialRequest
            },
            store: {
              common: {
                key: k2,
                signingKeyHash: signingKeyHash
              },
              expectedConfirmationToken: otherConfirmationToken,
              expectation: Snake.Crypto.FHMQV.MessageType.ConfirmationToken
            }
          };
        }.bind(this));
      },
      // TODO: fix this doc
      /**
       * Does the second phase of the FHMQV protocol.
       * @memberOf Snake.Crypto.FHMQV
       * @param {Snake.Model.Message} message message to update with the
       *        fields necessary to other end to proceed, namely
       *        confirmationToken. Note that in this case the recipientId
       *        field should not be set as it will be decrypted from the
       *        encryptedIdentity parameter.
       * @param {String} otherEphemeralPublicKey ephemeral public key chosen
       *        by the other end. Hex value.
       * @param {String} receivedOtherConfirmationToken confirmation token
       *        sent by the other end, will be compared with the expected
       *        one. Hex value.
       * @param {String} encryptedIdentity data containing the identity, i.e.
       *        the user ID, of the other end, encrypted with K0. Hex value.
       * @returns {Promise} a Promise that will be resolved when all the
       *          computations are done. The result of the promise is an hash
       *          with the fields to store in the friendship descriptor (key,
       *          the derived symmetric key, and expectedConfirmationToken).
       */
      // TODO: this is terrible and has to be fixed
      answer: function(receivedData) {
        // Return an object with the information to send/store
        return Snake.identityPromise({
          "send": {
            confirmationToken: receivedData._confirmationToken,
            type: Snake.Crypto.FHMQV.MessageType.ConfirmationToken
          },
          "store": {
            common: {
              key: receivedData._key,
              signingKeyHash: receivedData._signingKeyHash
            }
          },
          "done": true
        });
      },
      handleInitialRequest: function(receivedData) {
        var curve, myEphemeralKey, myEphemeralKeyJwk, myStaticKey,
            otherEphemeralPublicKey, otherEphemeralPublicKey,
            receivedOtherConfirmationToken, encryptedIdentity, other, otherId,
            myConfirmationToken, k2, otherEphemeralPublicKeyJwk;

        curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;

        receivedOtherConfirmationToken = receivedData.fhmqv.confirmationToken;
        encryptedIdentity = receivedData.fhmqv.identity;
        otherEphemeralPublicKey = curve.decodePointHex(
            util.abv2hex(receivedData.fhmqv.ephemeralPublicKey));

        // TODO: recycle ephemeral keys
        myEphemeralKey = new ECKey(Snake.Config.Crypto.curve,
            util.abv2hex(Snake.sessionUser.get("ephemeralPrivateKey")),
            util.abv2hex(Snake.sessionUser.get("ephemeralPublicKey")));
        myStaticKey = new ECKey(Snake.Config.Crypto.curve,
            util.abv2hex(Snake.sessionUser.get("staticPrivateKey")),
            util.abv2hex(Snake.sessionUser.get("staticPublicKey")));

        otherEphemeralPublicKey.validate();

        myEphemeralKeyJwk = {
          kty: "EC",
          crv: Snake.Config.Crypto.curve,
          d: util.b64encode(util.hex2abv(myEphemeralKey.getPriv().toHex())),
          x: util.b64encode(util.hex2abv(myEphemeralKey.getPubPoint().getX().toHex())),
          y: util.b64encode(util.hex2abv(myEphemeralKey.getPubPoint().getY().toHex()))
        };

        otherEphemeralPublicKeyJwk =
            Snake.Crypto.Common.ecAnsiPublicKeyToJwk(otherEphemeralPublicKey.serializeANSI());

        return Snake.Crypto.Common.deriveKey(myEphemeralKeyJwk,
                                             otherEphemeralPublicKeyJwk,
                                             ["decrypt"])
            .then(function(k0) {
              // TODO: put a real IV
              // Encrypt current user's ID using K0
              return Snake.Crypto.Common.symmAuthDecryption(k0,
                  encryptedIdentity, new Uint8Array(32));
            }.bind(this)).then(function(result) {
          otherId = parseInt(util.abv2hex(result), 16);
          other = new Snake.Model.User({id: otherId}, {});

          return other.fetch().result;
        }.bind(this)).then(function() {
          // TODO: ask the user if he wants to accept friendship
          var otherStaticPublicKey, fhmqv, otherConfirmationToken,
              signingKeyHashPromise;

          // Compute the hash of the signing key for storage
          // We'll wait for this promise later
          signingKeyHashPromise = Snake.Crypto.Common.hash(
              other.get("signingPublicKey"));

          otherStaticPublicKey = curve.decodePointHex(
              util.abv2hex(other.get("staticPublicKey")));
          otherStaticPublicKey.validate();

          // Derive the authenticated session key K2 and current
          // user's confirmation token
          fhmqv = new FHMQV(Snake.Config.Crypto.curve,
              CryptoJS.algo[Snake.Config.Crypto.hashAlgorithm.replace("-", "")],
              CryptoJS.algo[Snake.Config.Crypto.hmacAlgorithm],
              false /* initiator */,
              myEphemeralKey.getPriv(), myEphemeralKey.getPubPoint(),
              otherEphemeralPublicKey,
              myStaticKey.getPriv(), myStaticKey.getPubPoint(),
              otherStaticPublicKey,
              Snake.sessionUser.getDataToAuthenticate(),
              other.getDataToAuthenticate(),
              false /* validate */);

          otherConfirmationToken = fhmqv.getOtherConfirmationToken();
          if (!(util.memcmp(
              receivedOtherConfirmationToken, otherConfirmationToken))) {
            throw new Error("Confirmation token mismatch");
          }

          myConfirmationToken = fhmqv.getMyConfirmationToken();
          k2 = fhmqv.getK2();

          // Return signingKeyHashPromise to get the result of the previously
          // launched hash computation
          return signingKeyHashPromise;
        }.bind(this)).then(function(signingKeyHash) {
          _.extend(receivedData.fhmqv, {
            _confirmationToken: myConfirmationToken,
            _key: k2,
            _signingKeyHash: signingKeyHash
          });
          receivedData._senderId = otherId;
          receivedData._recipientType = "user";
          receivedData._recipientId = Snake.sessionUser.id;
        }.bind(this));
      },
      // TODO: fix this doc
      /**
       * Compares the received confirmation token with the expected one.
       * @memberOf Snake.Crypto.FHMQV
       * @param {String} confirmationToken the confirmation token in hex.
       * @param {Object} friendship friendship describing object to update
       *         in case of success.
       */
      checkConfirmationToken: function(receivedData, storedData) {
        var expectedConfirmationToken, actualConfirmationToken;
        expectedConfirmationToken = storedData.expectedConfirmationToken;
        actualConfirmationToken = receivedData.confirmationToken;

        if (!(util.memcmp(expectedConfirmationToken,
            actualConfirmationToken))) {
          throw new Error("Expected confirmation token (" +
              storedData.expectedConfirmationToken +
              ") does not match the received one (" +
              receivedData.confirmationToken + ").");
        }

        // Just keep the key
        return Snake.identityPromise({
          "done": true
        });
      }
    }, {
      "fieldName": "fhmqv"
    })
  });
}());
