(function() {
  "use strict";

  /**
   * Performs a rekey operation for one node.
   * @param {Obejct} rekeyOp
   * @param {Object} otherKey the new node where to store the new key.
   * @param {ArrayBufferView} iv the initialization vector.
   * @returns {Promise} resolved when the new key has been updated.
   */
  function applyRekey(rekeyOp, otherKey, iv) {
    var key = rekeyOp.oldKey.key;
    var data = rekeyOp.newEncKey.data;
    var newKeyId = rekeyOp.newEncKey.encKey;
    return Snake.Crypto.Common.symmetricDecryption(key, data, iv)
        .then(function(newKey) {
      otherKey.id = newKeyId;
      otherKey.key = newKey;
    });
  }

  /**
   * Returns a rekey opeartion to be supplied to the function applyRekey.
   * @param {Array} oldKeys array with old keys.
   * @param {Array} newEncKeys array with rekeying messages.
   * @returns {Object} a rekey operation object.
   */
  function getRekeyOperation(oldKeys, newEncKeys) {
    for (var i = 0; i < oldKeys.length; i++) {
      for (var j = 0; j < newEncKeys.length; j++) {
        if (oldKeys[i].id === newEncKeys[j].withKey) {
          var rekeyOp = {
            oldKey: oldKeys[i],
            newEncKey: newEncKeys[j]
          };
          newEncKeys.splice(j, 1);
          return rekeyOp;
        }
      }
    }
  }

  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    /**
     * @class Snake.Crypto.GroupProtocol
     */
    GroupProtocol: {
      /**
       * Handles the rekying message for the insertion of an user.
       * @param {Object} rekeyMessage the rekying message.
       * @param {Object} keys the keys to update.
       * @returns {Promise} resolved when the keys has been updated.
       * @memberOf Snake.Crypto.GroupProtocol#
       */
      handleInsertMessage: function(rekeyMessage, keys) {
        // An insert operation causes a rekey only if it is created a new root.
        if (rekeyMessage.rootKey && rekeyMessage.newMember !== Snake.sessionUser.id) {
          // The old root key becomes an auxiliary key
          keys.auxiliaryKeys.push(_.clone(keys.rootKey));
          // Updates the root key using the single rekey message for the root node.
          var rekeyOp = {
            oldKey: keys.rootKey,
            newEncKey: rekeyMessage.rootKey
          };
          return applyRekey(rekeyOp, keys.rootKey, rekeyMessage.iv);
        } else {
          return Snake.identityPromise();
        }
      },
      /**
       * Handles the rekying message for the deletion of an user.
       * @param {Object} rekeyMessage the rekying message.
       * @param {Object} keys the keys to update.
       * @returns {Promise} resolved when the keys has been updated.
       * @memberOf Snake.Crypto.GroupProtocol#
       */
      handleRemoveMessage: function(rekeyMessage, keys) {
        var newEncKeys = rekeyMessage.keys;
        var iv = rekeyMessage.iv;

        // TODO check if this blows up using native promises
        var rekeyCicle = function(rekeyOp) {
          if(!rekeyOp) {
            rekeyOp = getRekeyOperation(newAuxiliaryKeys, newEncKeys);
          }

          if (rekeyOp) {
            var newKey = {};
            newAuxiliaryKeys.push(newKey);
            return applyRekey(rekeyOp, newKey, iv).then(rekeyCicle);
          } else {
            return Snake.identityPromise();
          }
        };

        var newAuxiliaryKeys = [];
        var rekeyOp;
        do {
          rekeyOp = getRekeyOperation([keys.auxiliaryKeys[0]], newEncKeys);
          newAuxiliaryKeys.push(keys.auxiliaryKeys.shift());
        } while (!rekeyOp);

        return rekeyCicle(rekeyOp).then(function() {
          keys.rootKey = newAuxiliaryKeys.pop();
          keys.auxiliaryKeys = newAuxiliaryKeys;
        });
      },
      /**
       * Inserts a member in a group. The function sends the rekeying message
       * both to the new user and to the group.
       * @param {Integer} newUserId the id of the user to insert.
       * @param {Object} group the model of the group.
       * @returns {Promise} Promise resolved when all the messages are sent.
       * @memberOf Snake.Crypto.GroupProtocol#
       */
      insertMember: function(newUserId, newUserKey, group) {
        var bush = group.adminData.bush;
        var rekeyMessage, groupRekeyPost, userIdMac, groupIdMac;

        // TODO lock the group before performing the rekey

        // TODO For the time being the adminstrator sends a private message to
        //      the user that wants to add to the group. This requires that
        //      the new user and the administrator are friends.
        // Inserts the user into the group tree
        return bush.insert(newUserId).then(function(res) {
          rekeyMessage = res;

          // Sends a message to the group that notifies about the new user
          groupRekeyPost = new Snake.Model.Post({
            _recipientType: "group",
            _recipientId: group.id,
            authorId: Snake.sessionUser.id,
            type: "member-added",
            group: {
              addedBy: Snake.sessionUser.id,
              newMember: newUserId,
              rekey: rekeyMessage.other
            }
          }, {});
          return groupRekeyPost.save().result;
        }).then(function() {
          // XXX lastValidMessageId is not always the rekeying post because
          //     an insert rekey usually does not change the group key. Try to
          //     find something more elegant than accessing directly Snake.sessionUser
          var lastValidMessageId = rekeyMessage.rootKey ? groupRekeyPost.id :
              _(Snake.sessionUser.get("groups")).findByField(group.id).lastValidMessageId;

          // Sends a message to the new user with all its keys
          var userRekeyMessage = new Snake.Model.Message({
            _recipientType: "user",
            _recipientId: newUserId,
            _senderId: Snake.sessionUser.id,
            type: "group-joined",
            group: {
              addedBy: Snake.sessionUser.id,
              groupId: group.id,
              address: group.get("address"),
              keys: rekeyMessage.newUser,
              lastValidMessageId: lastValidMessageId
            }
          }, {});
          return userRekeyMessage.save().result;
        }).then(function() {
          // Processes the rekeying message for the group, it is necessary to
          // wait this because the descriptor of the group needs to be encrypted
          // with the new group key
          return group.processNewMessages();
        }).then(function() {
          // If the group keys is changed save the old one in the group descriptor
          if (rekeyMessage.rootKey) {
            group.get("oldKeys").push({
              key: rekeyMessage.oldRootKey,
              lastValidMessageId: groupRekeyPost.id
            });
          }
          // Inserts the user into the members of the group
          var members = group.get("members");
          members.push({id: newUserId, staticPublicKey: newUserKey});
          group.setChanged("members");
          var groupPromise = group.save().result;

          // Updates the administrative data of the group
          group.adminData.serializeBush();
          var groupAdminPromise = group.adminData.save().result;

          return Promise.every(groupPromise, groupAdminPromise);
        }).then(function() {
          // TODO unlock the group once that the rekeying is done
          return group.getEncryptionKey();
        }).then(function(groupKey){
          // Once we rekey, let's add the user to the web of trust.
          // Calculate the HMACs of both the user id and the group id
          userIdMac = Snake.Crypto.Common
            .mac(groupKey, util.hex2abv(newUserId.toString(16)));
          groupIdMac = Snake.Crypto.Common
            .mac(groupKey, util.hex2abv(group.id.toString(16)));
          return Promise.every(userIdMac, groupIdMac).then(function(macs) {
            var entry = new Snake.Model.WoTEntry({
              userIdMac: macs[0],
              groupIdMac: macs[1],
              editToken: new Uint8Array(1) // bogus data
            }, {});

            return entry.save().result;
          });
        });
      },
      /**
       * Removes a member from a group. The function sends the rekeying message
       * both to the old user and to the group.
       * @param {Integer} userToRemove the id of the user to remove.
       * @param {Object} group the model of the group.
       * @returns {Promise} Promise resolved when all the messages are sent.
       * @memberOf Snake.Crypto.GroupProtocol#
       */
      removeMember: function(userToRemove, group) {
        var bush = group.adminData.bush;
        var rekeyMessage, groupRekeyPost;

        // TODO lock the group before performing the rekey

        // TODO For the time being the adminstrator sends a private message to
        //      the user that wants to remove from the group. This requires that
        //      the new user and the administrator are friends.
        // Removes the user from the group tree
        return bush.remove(userToRemove).then(function(res) {
          rekeyMessage = res;

          // Sends a message to the group that notifies about the removed user
          groupRekeyPost = new Snake.Model.Post({
            _recipientType: "group",
            _recipientId: group.id,
            authorId: Snake.sessionUser.id,
            type: "member-removed",
            group: {
              removedBy: Snake.sessionUser.id,
              oldMember: userToRemove,
              rekey: rekeyMessage
            }
          }, {});
          return groupRekeyPost.save().result;
        }).then(function() {
          // Sends a message to the old user to notify that he has been removed.
          var userRekeyMessage = new Snake.Model.Message({
            _recipientType: "user",
            _recipientId: userToRemove,
            _senderId: Snake.sessionUser.id,
            type: "group-leaved",
            group: {
              removedBy: Snake.sessionUser.id,
              groupId: group.id
            }
          }, {});
          return userRekeyMessage.save().result;
        }).then(function() {
          // Processes the rekeying message for the group, it is necessary to
          // wait this because the descriptor of the group needs to be encrypted
          // with the new group key
          return group.processNewMessages();
        }).then(function() {
          // Saves the old group key in the group descriptor
          group.get("oldKeys").push({
            key: rekeyMessage.oldRootKey,
            lastValidMessageId: groupRekeyPost.id
          });
          // Removes the user from the members of the group
          var members = group.get("members");
          members.splice(_.indexOf(members, _.find(members, function(member) {
            return member.id === userToRemove;
          })), 1);
          group.setChanged("members");
          var groupPromise = group.save().result;

          // Updates the administrative data of the group
          group.adminData.serializeBush();
          var groupAdminPromise = group.adminData.save().result;

          return Promise.every(groupPromise, groupAdminPromise);
        }).then(function() {
          // TODO unlock the group once that the rekeying is done
        });
      },
      promoteMemberToAdministrator: function(userToPromote, group) {
          var groupKeys = _.findByField(Snake.sessionUser.get("groups"), group.id);

          // Sends a message to the user that notifies about the removed user
          var adminGroupMessage = new Snake.Model.Message({
            _recipientType: "user",
            _recipientId: userToPromote,
            _senderId: Snake.sessionUser.id,
            type: "group-admin-access",
            group: {
              groupId: group.id,
              promotedBy: Snake.sessionUser.id,
              signingPrivateKey: groupKeys.signingPrivateKey,
              adminKey: groupKeys.adminKey
            }
          }, {});
          return adminGroupMessage.save().result;
      }
    }

  });
}());
