// TODO: salt dataToAuthenticate, otherwise shared secrets are equal if inputs
//       are
// TODO: decouple UI using callbacks
// TODO: decouple user info fetch

(function() {
  "use strict";

  // Support functions
  function abv2point(curve, abv) {
    return curve.decodePointHex(util.abv2hex(abv));
  }

  function pubkey2abv(key) {
    return util.hex2abv(key.getPubPoint().encodeHex());
  }

  function abv2eckey(abv) {
    return new ECKey(Snake.Config.Crypto.curve, util.abv2hex(abv));
  }

  function computeSharedSecret(secret, dataToAuthenticate) {
    return Snake.Crypto.Common.macWithHashedKey(util.utf8encode(secret),
        dataToAuthenticate).then(function(result) {
      Snake.Logging.log("friendship", "Shared secret computed: " +
          util.abv2hex(result));
      return result;
    });
  }

  Snake.Crypto = _.extend(Snake.Crypto || {}, {
    SMP: _(_(Snake.Crypto.FriendshipProtocol).clone()).extend({
      MessageType: {
        "Step1": 0,
        "Step2": 1,
        "Step3": 2,
        "Step4": 3
      },
      MessageHandler: [
        "step2",
        "step3",
        "step4",
        "step5"
      ],
      // From A to B
      step1: function(otherId, isAuthenticated) {
        var toSend, toStore, other, smpAB, smpBA, keys;

        // If the user has already authenticated the friend with an offline
        // mechanism then return immediately from the SMP with a positive result
        if (isAuthenticated) {
          return {
            "store": null,
            "send": null,
            "done": true
          };
        }

        toSend = {type: Snake.Crypto.SMP.MessageType.Step1};
        toStore = {expectation: Snake.Crypto.SMP.MessageType.Step2};

        // Generate the needed key pairs
        keys = SMP.generateInitialKeys(Snake.Config.Crypto.curve);
        toSend.publicKey2 = pubkey2abv(keys.key2);
        toSend.publicKey3 = pubkey2abv(keys.key3);
        toStore.privateKey2 = util.hex2abv(keys.key2.getPriv().toHex());
        toStore.privateKey3 = util.hex2abv(keys.key3.getPriv().toHex());

        // Start by fetching other user's details
        other = new Snake.Model.User({id: otherId}, {});

        return other.fetch().result.then(function() {
          return Snake.Crypto.SMP.askQuestion(other.get("username"));
        }.bind(this)).then(function(userQuestion) {
          // Check if the user wants to skip public key authentication
          if (_(userQuestion).isUndefined()) {
            // Return immediately, no delay necessary
            return Snake.identityPromise({
                send: toSend,
                store: toStore
            });
          } else {
            toSend.aToB = {};
            toStore.aToB = {};

            smpAB = new SMP(Snake.Config.Crypto.curve, true);
            smpAB.myKey2 = keys.key2;
            smpAB.myKey3 = keys.key3;

            toSend.aToB.question = userQuestion.myQuestion;

            return computeSharedSecret(userQuestion.myAnswer,
                Snake.concatBuffers(Snake.sessionUser.getDataToAuthenticate(),
                other.getDataToAuthenticate()))
                .then(function(secret) {
              // Save the result for the next step so we don't have
              // to ask the answer later again
              toStore.aToB.sharedSecret = secret;

              return {
                "store": toStore,
                "send": toSend
              };
            });
          }
        });
      },
      // From B to A
      step2: function(receivedData, storedData, otherId) {
        var curve, smpAB, other, toSend, toStore, dataToAuthenticate,
            smpBA, otherPublicKey2, otherPublicKey3, keys;

        curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;
        toSend = {type: Snake.Crypto.SMP.MessageType.Step2};
        toStore = {expectation: Snake.Crypto.SMP.MessageType.Step3};

        // Generate the needed key pairs
        keys = SMP.generateInitialKeys(Snake.Config.Crypto.curve);
        toSend.publicKey2 = pubkey2abv(keys.key2);
        toSend.publicKey3 = pubkey2abv(keys.key3);
        // We do not store key2, it is not used in step4
        // toStore.privateKey2 = util.hex2abv(keys.key2.getPriv().toHex());
        toStore.privateKey3 = util.hex2abv(keys.key3.getPriv().toHex());

        // Get the needed public keys
        otherPublicKey2 = abv2point(curve, receivedData.publicKey2);
        otherPublicKey2.validate();
        otherPublicKey3 = abv2point(curve, receivedData.publicKey3);
        otherPublicKey3.validate();

        return otherId.then(function(senderId) {
          // Start by fetching other user's details
          other = new Snake.Model.User({id: senderId}, {});

          return other.fetch().result;
        }).then(function() {
          // TODO: accept/refuse here?

          dataToAuthenticate = Snake.concatBuffers(
              other.getDataToAuthenticate(),
              Snake.sessionUser.getDataToAuthenticate());

          // Check if the other end intends to authenticate me
          if (!_(receivedData).has("aToB")) {
            // He seems to trust our public key, do we?
            return Snake.Crypto.SMP.askQuestion(other.get("username"));
          }

          // The other use wants to authenticate our public key
          smpAB = new SMP(Snake.Config.Crypto.curve, false);
          smpAB.myKey2 = keys.key2;
          smpAB.myKey3 = keys.key3;
          smpAB.otherPublicKey2 = otherPublicKey2;
          smpAB.otherPublicKey3 = otherPublicKey3;

          // Let's ask the user the answer and if he wants to
          // counter-authenticate the other user
          return Snake.Crypto.SMP.answerAndAskBack(other.get("username"),
                                                   receivedData.aToB.question)
              .then(function(userAnswer) {
            var resultAB, resultBA;

            if (_(userAnswer).isUndefined()) {
              // TODO: better handling
              throw new Error("Operation canceled.");
            }

            resultAB = Snake.identityPromise();
            resultBA = Snake.identityPromise();

            // Are we answering A's question?
            if (!_(userAnswer.otherAnswer).isUndefined()) {
              resultAB =
                computeSharedSecret(userAnswer.otherAnswer,
                  dataToAuthenticate).then(function(secret) {

                smpAB.sharedSecret = new BigInteger(util.abv2hex(secret), 16);

                toStore.aToB = {
                  p: util.hex2abv(smpAB.getMyP().encodeHex()),
                  q: util.hex2abv(smpAB.getMyQ().encodeHex())
                };

                toSend.aToB = {
                  p: util.hex2abv(smpAB.getMyP().encodeHex()),
                  q: util.hex2abv(smpAB.getMyQ().encodeHex())
                };
              });
            }

            // Do we want to authenticate the other end?
            if (userAnswer.unsafe === "on") {
              smpBA = new SMP(Snake.Config.Crypto.curve, true);
              smpBA.myKey2 = keys.key2;
              smpBA.myKey3 = keys.key3;
              smpBA.otherPublicKey2 = otherPublicKey2;
              smpBA.otherPublicKey3 = otherPublicKey3;

              resultBA = computeSharedSecret(userAnswer.myAnswer,
                dataToAuthenticate).then(function(secret) {

                smpBA.sharedSecret = new BigInteger(util.abv2hex(secret), 16);

                // Save only the shared secret, private keys have already been
                // saved
                toStore.bToA = {
                  p: util.hex2abv(smpBA.getMyP().encodeHex()),
                  q: util.hex2abv(smpBA.getMyQ().encodeHex()),
                  sharedSecret: secret
                };

                toSend.bToA = {
                  question: userAnswer.myQuestion,
                  p: util.hex2abv(smpBA.getMyP().encodeHex()),
                  q: util.hex2abv(smpBA.getMyQ().encodeHex())
                };

              });
            }

            return Promise.every(resultAB, resultBA);
          });
        }).then(function() {
          return {
            "store": toStore,
            "send": toSend
          };
        });
      },
      // From A to B
      step3: function(receivedData, storedData, otherId) {
        var curve, smpAB, smpBA, toSend, toStore, keys, otherPublicKey2,
	    otherPublicKey3, other, dataToAuthenticate;

        curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;
        toSend = {type: Snake.Crypto.SMP.MessageType.Step3};
        toStore = {expectation: Snake.Crypto.SMP.MessageType.Step4};

        keys = {};
        keys.key2 = abv2eckey(storedData.privateKey2);
        keys.key3 = abv2eckey(storedData.privateKey3);
        // We'll need key3 in step5, forward it
        toStore.privateKey3 = storedData.privateKey3;

        otherPublicKey2 = abv2point(curve, receivedData.publicKey2);
        otherPublicKey2.validate();
        otherPublicKey3 = abv2point(curve, receivedData.publicKey3);
        otherPublicKey3.validate();

        // A -> B
        if (_(storedData).has("aToB")) {
          smpAB = new SMP(Snake.Config.Crypto.curve, true);
          smpAB.otherPublicKey2 = otherPublicKey2;
          smpAB.otherPublicKey3 = otherPublicKey3;

          smpAB.sharedSecret =
              new BigInteger(util.abv2hex(storedData.aToB.sharedSecret), 16);
          smpAB.myKey2 = keys.key2;
          smpAB.myKey3 = keys.key3;
          smpAB.otherQ = abv2point(curve, receivedData.aToB.q);

          toSend.aToB = {
            p: util.hex2abv(smpAB.getMyP().encodeHex()),
            q: util.hex2abv(smpAB.getMyQ().encodeHex()),
            r: util.hex2abv(smpAB.getMyR().encodeHex())
          };

          toStore.aToB = {
            p: util.hex2abv(smpAB.getMyP().encodeHex()),
            otherP: receivedData.aToB.p
          };
        }

        // B -> A
        // If B sent us a message with bToA, he wants to counterauthenticate us
        if (_(receivedData).has("bToA")) {
          smpBA = new SMP(Snake.Config.Crypto.curve, false);
          smpBA.otherPublicKey2 = otherPublicKey2;
          smpBA.otherPublicKey3 = otherPublicKey3;
          smpBA.myKey2 = keys.key2;
          smpBA.myKey3 = keys.key3;

          smpBA.otherQ = abv2point(curve, receivedData.bToA.q);

          return otherId.then(function(senderId) {
            // Start by fetching other user's details
            other = new Snake.Model.User({id: senderId}, {});

            return other.fetch().result;
          }).then(function() {
            return Snake.Crypto.SMP.answerAndAskBack(other.get("username"),
                                                     receivedData.bToA.question);
          }).then(function(userAnswer) {
            dataToAuthenticate = Snake.concatBuffers(
              Snake.sessionUser.getDataToAuthenticate(),
              other.getDataToAuthenticate());

            return computeSharedSecret(userAnswer.otherAnswer,
              dataToAuthenticate).then(function(secret) {

              smpBA.sharedSecret = new BigInteger(util.abv2hex(secret), 16);

              toSend.bToA = {
                p: util.hex2abv(smpBA.getMyP().encodeHex()),
                q: util.hex2abv(smpBA.getMyQ().encodeHex()),
                r: util.hex2abv(smpBA.getMyR().encodeHex())
              };

              toStore.bToA = {
                p: util.hex2abv(smpBA.getMyP().encodeHex()),
                otherP: receivedData.bToA.p
              };

              return Snake.identityPromise({
                "store": toStore,
                "send": toSend
              });

            });
          });
        } else {
          return Snake.identityPromise({
            "store": toStore,
            "send": toSend
          });
        }
      },
      // From B to A
      step4: function(receivedData, storedData) {
        var curve, smpAB, smpBA, toSend, toStore, keys;

        curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;
        toSend = {type: Snake.Crypto.SMP.MessageType.Step4};
        toStore = null;

        keys = {};
        keys.key3 = abv2eckey(storedData.privateKey3);

        // A -> B
        if (_(storedData).has("aToB")) {
          smpAB = new SMP(Snake.Config.Crypto.curve, false);

          smpAB.myKey3 = keys.key3;
          smpAB.myP = abv2point(curve, storedData.aToB.p);
          smpAB.myQ = abv2point(curve, storedData.aToB.q);

          smpAB.otherP = abv2point(curve, receivedData.aToB.p);
          smpAB.otherQ = abv2point(curve, receivedData.aToB.q);
          smpAB.otherR = abv2point(curve, receivedData.aToB.r);

          if (!smpAB.checkResult()) {
            throw new Error("Shared secrets do not match");
          }

          toSend.aToB = {
            r: util.hex2abv(smpAB.getMyR().encodeHex())
          };
        }

        // B -> A
        if (_(storedData).has("bToA")) {
          smpBA = new SMP(Snake.Config.Crypto.curve, true);

          smpBA.myKey3 = keys.key3;
          smpBA.myP = abv2point(curve, storedData.bToA.p);
          smpBA.myQ = abv2point(curve, storedData.bToA.q);

          smpBA.otherP = abv2point(curve, receivedData.bToA.p);
          smpBA.otherQ = abv2point(curve, receivedData.bToA.q);
          smpBA.otherR = abv2point(curve, receivedData.bToA.r);

          if (!smpBA.checkResult()) {
            throw new Error("Shared secrets do not match");
          }

          toSend.bToA = {
            r: util.hex2abv(smpBA.getMyR().encodeHex())
          };
        }

        return Snake.identityPromise({
          "store": toStore,
          "send": toSend,
          "done": true
        });
      },
      // From A to B
      step5: function(receivedData, storedData) {
        var curve, smpAB, smpBA, toSend, toStore, keys;

        curve = getSECCurveByName(Snake.Config.Crypto.curve).curve;
        toSend = null;
        toStore = {
          "common": {
            "isAuthenticated": true
          }
        };

        keys = {};
        keys.key3 = abv2eckey(storedData.privateKey3);

        // A -> B
        if (_(storedData).has("aToB")) {
          smpAB = new SMP(Snake.Config.Crypto.curve, true);

          smpAB.myKey3 = keys.key3;
          smpAB.myP = abv2point(curve, storedData.aToB.p);
          smpAB.otherP = abv2point(curve, storedData.aToB.otherP);

          smpAB.otherR = abv2point(curve, receivedData.aToB.r);

          if (!smpAB.checkResult()) {
            throw new Error("Shared secrets do not match");
          }
        }

        // B -> A
        if (_(storedData).has("bToA")) {
          smpBA = new SMP(Snake.Config.Crypto.curve, false);

          smpBA.myKey3 = keys.key3;
          smpBA.myP = abv2point(curve, storedData.bToA.p);
          smpBA.otherP = abv2point(curve, storedData.bToA.otherP);

          smpBA.otherR = abv2point(curve, receivedData.bToA.r);

          if (!smpBA.checkResult()) {
            throw new Error("Shared secrets do not match");
          }
        }

        return Snake.identityPromise({
          "store": toStore,
          "send": toSend,
          "done": true
        });
      }
    }, {
     "fieldName": "smp"
    })
  });
}());
