(function() {
  "use strict";

  var Crypto = Snake.Crypto;
  var OfflineAuthenticator = Crypto.OfflineAuthenticator;

  _.extend(Crypto, {}, {
    /**
     * Public key authenticator using the Web of Trust to verify the validity of
     * a public key.
     * In practice, this method looks in the friend list of all the current user's
     * friends to check if more of fixed threshold of them have seen the same
     * public key.
     *
     * @class Snake.Crypto.WebOfTrust
     * @augments Snake.Crypto.OfflineAuthenticator
     */
    WebOfTrust: _(_(Crypto.OfflineAuthenticator).clone()).extend({
      name: "WebOfTrust",
      verify: function(id) {
        var newFriend, groups, wotData, entries, commonGroups, survivors,
            groupPromises, candidates;

        groups = Snake.sessionUser.get("groups");
        newFriend = new Snake.Model.User({id: id}, {});
        // Temp structures for keeping the data of this authentication run
        wotData = [];
        candidates = [];

        // Fetch new friend's details
        return newFriend.fetch().result.then(function() {
          // For each group derive the HMAC the newFriend ID and the group ID
          // using the group key as key
          groupPromises = groups.map(function(group){
            return Snake.Model.Group.getEncryptionKey({id: group.id})
            .then(function(groupKey){
              return Promise.every(
                Crypto.Common.mac(groupKey, util.hex2abv(id.toString(16))),
                Crypto.Common.mac(groupKey, util.hex2abv(group.id.toString(16)))
              );
            }).then(function(macs){
              wotData.push({
                newFriendIdMac: macs[0],
                groupIdMac: macs[1],
                id: group.id
              });
            });
          });
          return Promise.all(groupPromises);
        }).then(function() {
          // Fetch the proper WoT entries by newFriendIdMac
          entries = new Snake.Model.Collection([], {
            model: Snake.Model.WoTEntry,
            url: function() {
              return {
                object: "common-groups",
                members: _(wotData).pluck("newFriendIdMac")
              };
            }.bind(this)
          });
          return entries.fetch().result;
        }.bind(this)).then(function() {
          // Filter out undefined values and keep only valid results
          // that came from the server (the `entries` array).
          // To do that we sort both our data container and the array of entries
          // by the groupIdMac, and ignore the groups we kept that are not
          // the common friends groups.
          var i = 0;

          wotData = _(wotData).sortBy(function(row){ return row.groupIdMac; });

          _(entries.pluck("groupIdMac").sort()).each(function(entry){
            while(!util.memcmp(wotData[i].groupIdMac, entry)){ i++; }
            candidates.push(wotData[i]);
          });

          // Drop early if we have less friends in common than
          // the established threshold
          if (candidates.length < Snake.Config.WebOfTrust.degrees) {
            return OfflineAuthenticator.AuthResult.Unknown;
          } else {
            // Now that we have the common groups we can actually fetch the
            // groups data for public keys checking
            commonGroups = new Snake.Model.Collection([], {
              model: Snake.Model.Group,
              url: function() {
                return {
                  object: "group",
                  ids: _(candidates).pluck("id")
                };
              }.bind(this)
            });

            return commonGroups.fetch().result.then(function() {
              // Finally match the PublicKey of the new friend!
              // Here we filter because it may be that the server lied in giving
              // us the data; once we're done with the checks we count if we
              // still have a sufficient number of survivors to proceed
              survivors = commonGroups.filter(function(group) {
                // .some returns true if function is true at least once.
                // since we're testing also the id we're sure it's EXACTLY once
                return group.get("members").some(function(member){
                  return util.memcmp(newFriend.get("staticPublicKey"),
                                     member.staticPublicKey) &&
                         newFriend.get("id") === member.id;
                });
              });

              if (survivors.length < Snake.Config.WebOfTrust.degrees) {
                Snake.Logging.warn("friendship",
                          "Public key mismatch while using the Web of Trust!");
                return OfflineAuthenticator.AuthResult.Unknown;
              }
              Snake.Logging.log("friendship",
                                "Friendship established using Web of Trust!");
              return OfflineAuthenticator.AuthResult.Valid;
            }.bind(this));
          }
        }.bind(this));
      }
    })
  });
}());
