(function() {
  "use strict";

  var tables;

  function insert(tableName, payload) {
    var now = new Date().toISOString();
    var table = tables[tableName];
    var toSave = _.clone(payload);
    toSave.id = table.counter++;
    toSave.date = now;

    if (_.has(table.data, toSave.id))
      throw new Error("Duplicate key");

    table.data[toSave.id] = toSave;

    return {
      status: "Ok",
      id: toSave.id,
      date: now
    };
  }

  function update(tableName, payload) {
    var now = new Date().toISOString();
    var table = tables[tableName];

    if (!_.has(table.data, payload.id)) {
      throw new Error("The requested ID " + payload.id.toString() +
          " was not found in table \"" + tableName + "\"");
    }

    var toSave = _.clone(payload);
    toSave.date = now;
    _.extend(table.data[toSave.id], toSave);

    return {
      status: "Ok",
      id: toSave.id,
      date: now
    };
  }

  function read(tableName, ids, exclude) {
    if (_.isUndefined(exclude)) {
      exclude = [];
    }

    var table = tables[tableName];

    var find = function (id) {
      var found = _.findByField(table.data, id, "id");
      if (_.isUndefined(found)) {
        throw new Error("The requested ID " + id.toString() +
            " was not found in table \"" + tableName + "\"");
      }

      return _.omit(_.clone(found), exclude);
    };

    return _.map(ids, find);
  }

  function readBy(tableName, field, id) {
    var table = tables[tableName];

    var matching = _.findByField(table.data, id, field);

    if (_.isUndefined(matching)) {
      return [];
    }

    return _.clone(matching);
  }

  function messagesAfterId(addresses, num) {
    var table = tables.messages;

    var messageFilter = function (message) {
        return _.reduce(addresses, function (prev, entry) {
            return prev ||
                message.address === entry.address &&
                message.id > entry.id;
        }, false);
    };

    return _.map(_.last(_.filter(table.data, messageFilter), num), _.clone);
  }

  function lastMessages(addresses, num) {
    var table = tables.messages;

    var messageFilter = function (entry) {
      return _.contains(addresses, entry.address);
    };

    var matching = _.filter(table.data, messageFilter);

    return _.last(_.map(_.sortBy(matching, "id"), _.clone), num);
  }

  Snake.LocalNetwork = {
    init: function () {
      // Replace the real network layer with the local one
      Snake.Network = Snake.LocalNetwork;

      Snake.tables = tables = {
        "users": {
          counter: 0,
          data: []
        },
        "profiles": {
          counter: 0,
          data: []
        },
        "groups": {
          counter: 0,
          data: []
        },
        "messages": {
          counter: 0,
          data: []
        },
        "wot": {
          counter: 0,
          data: []
        }
      };
    },
    read: function (payload) {
      return Snake.identityPromise().then(function () {
        var result;
        payload = _.clone(payload);
        var type = payload.object;
        delete payload.object;

        switch (type) {
          case "session-user":
            return read("users", payload.ids);
          case "user":
            return read("users", payload.ids, ["privateData"]);
          case "profile":
            return read("profiles", payload.ids);
          case "group":
            return read("groups", payload.ids, ["adminData"]);
          case "group-admin":
            result = read("groups", payload.ids, ["privateData"]);
            return _.map(result, function (a) {
              a.privateData = a.adminData;
              delete a.adminData;
              return a;
            });
          case "message":
            return read("messages", payload.ids);
          case "username":
            return [readBy("users", "username", payload.username)];
          case "messages-after-id":
            return messagesAfterId(payload.addresses, payload.num);
          case "last-messages":
            Snake.Logging.log("network", lastMessages(payload.addresses, payload.num));
            return lastMessages(payload.addresses, payload.num);
          case "common-groups":
            return _(_.map(payload.members, function (m) {
              result = readBy("wot", "userIdMac", m);
              delete result.editToken;
              delete result.userIdMac;
              delete result.privateData;
              return result;
            })).filter(function(row){return !_(row).isEmpty(); });
          default:
            throw new Error("\"" + type + "\" request not implemented");
        }
      });
    },
    create: function (payload) {
      return Snake.identityPromise().then(function () {
        var type = payload.object;
        delete payload.object;

        switch (type) {
          case "session-user":
            return insert("users", payload);
          case "profile":
            return insert("profiles", payload);
          case "group":
            return insert("groups", payload);
          case "message":
            return insert("messages", payload);
          case "wot-entry":
            return insert("wot", payload);
          default:
            throw new Error("\"" + type + "\" request not implemented");
        }
      });
    },
    update: function (payload) {
      return Snake.identityPromise().then(function () {
        var type = payload.object;
        delete payload.object;

        switch (type) {
          case "session-user":
            return update("users", payload);
          case "profile":
            return update("profiles", payload);
          case "group":
            return update("groups", payload);
          case "group-admin":
            payload.adminData = payload.privateData;
            delete payload.privateData;
            return update("groups", payload);
          case "message":
            return update("messages", payload);
          default:
            throw new Error("\"" + type + "\" request not implemented");
        }
      });
    },
    reset: function (payload) {
      return Snake.identityPromise().then(function () {
        var type = payload.object;
        delete payload.object;

        switch (type) {
          case "clean":
            return this.init();
          default:
            throw new Error("\"" + type + "\" request not implemented");
        }
      }.bind(this));
    }
  };
}());
