(function() {
  "use strict";

  var Crypto = Snake.Crypto;

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for a group.
     * @class Snake.Model.Group
     */
    Group: Snake.Model.ModelBase.extend({
      childModels: ["administrators", "members", "messages", "friendsOf"],
      defaultValues: {
        name: "",
        description: "",
        administrators: [],
        members: [],
        oldKeys: []
      },
      initialize: function(attributes, options) {
        this.options = _.defaults(options || {}, {
          setDefaultValues: false,
          fetchMembers: false,
          fetchAdministrators: false,
          fetchAdminData: false,
          fetchMessages: 0,
          fetchFriendsOf: false,
          members: {},
          administrators: {},
          messages: {},
          friendsOf: {}
        });

        // Copies defaults values into the model
        if (this.options.setDefaultValues) {
          this.loadDefaultValues();
        }

        // HACK: if there are collection options, merge them into the model opts
        if (_.has(this, "collection")) {
          this.options = _.extend(this.options, this.collection.options);
        }

        if (this.options.fetchMembers) {
          this.members = new Snake.Model.Collection([], {
            model: Snake.Model.User,
            url: function() {
              if (!this.has("members") || this.get("members").length === 0)
                return;

              // Set the query for the server, all the user IDs
              return {
                object: "user",
                ids: _(this.get("members")).pluck("id")
              };
            }.bind(this)
          });

          this.monitorChange("members");
        }

        if (this.options.fetchAdministrators) {
          this.administrators = new Snake.Model.Collection([], {
            model: Snake.Model.User,
            url: function() {
              if (!this.has("administrators") ||
                  this.get("administrators").length === 0) return;

              // Set the query for the server, all the administrato IDs
              return {
                object: "user",
                ids: _(this.get("administrators")).pluck("id")
              };
            }.bind(this)
          });

          this.monitorChange("administrators");
        }

        if (this.options.fetchAdminData) {
          this.adminData = new Snake.Model.GroupAdmin({id: this.id});
          // HACK adminData is not an attribute of the group model, by
          //      monitoring on changes of the attribute date adminData will
          //      be fetched automatically
          this.monitorChange("adminData", "date");

          // TODO bind on changes of adminData:date
        }

        if (this.options.fetchMessages > 0) {
          this.messages = new Snake.Model.Collection([], {
            model: Snake.Model.Post,
            comparator: Snake.dateComparator(),
            url: function() {
              if (!this.has("id")) return;
              return {
                object: "last-messages",
                addresses: [this.get("address")],
                num: this.options.fetchMessages
              };
            }.bind(this)
          });

          this.monitorChange("messages", "address");
        }

        if (this.options.fetchFriendsOf) {
          this.friendsOf = new Snake.Model.User({}, this.options.friendsOf);

          this.monitorChange("friendsOf");
        }
      },
      /**
       * Creates a new group for the SessionUser with the given in data.
       * At the end of group is initialized, saved and added to the sessionUser.
       * @memberOf Snake.Model.Group#
       * @param {Object} groupData: the object that contains the group data.
       *                            mandatory keys: "name", "description"
       * @returns {Promise} Promise resolved when all the crypto is done
       */
      setupNew: function(groupData) {
        // Creates model its administrative data
        var groupAdmin = new Snake.Model.GroupAdmin({}, {setDefaultValues: true});
        // Data unpacking
        if (_.isUndefined(groupData.name) || _.isUndefined(groupData.description))
          throw new Error("groupData should have atleast the 'name' and " +
                          "'description' properties");
        var name = groupData.name;
        var description = groupData.description;

        // Generates the symmetric key used to encrypt the administrative data,
        // the asymmetric key pair of the group and initialized the bush to
        // create the group key
        var publicGroupKey, privateGroupKey, groupKeys, groupAdminKey;
        return Snake.sessionUser.getAuthenticationToken().then(function(authTok){
          this.get("administrators").push({id: Snake.sessionUser.id,
                                           publicKey: authTok});
        }.bind(this)).then(function(){
          return Crypto.Common.generateSymmAuthKey();
        }.bind(this)).then(function(key) {
          groupAdminKey = key;
          groupAdmin.encryptionKey = key;
          return Crypto.Common.generateSigningKeyPair();
        }.bind(this)).then(function(keyPair) {
          publicGroupKey = Crypto.Common.jwkToAnsiPublicKey(keyPair.publicKey);
          this.set("signingPublicKey", publicGroupKey);
          privateGroupKey = Crypto.Common.jwkToPrivateKey(keyPair.privateKey);
          this.signingPrivateKey = privateGroupKey;
          return groupAdmin.initializeBush(Snake.sessionUser.id);
        }.bind(this)).then(function(keys) {
          groupKeys = keys;
          this.encryptionKey = keys.rootKey.key;
          this.set("name", name);
          this.set("description", description);
          this.set("address", Snake.generateAddress());
          return this.save().result;
        }.bind(this)).then(function(){
          // Saves the administrative data of the group
          groupAdmin.set("id", this.id);
          groupAdmin.set("date", this.get("date"));
          groupAdmin.save();

          // Saves the new group into the sessionUser's groups
          var groups = Snake.sessionUser.get("groups");
          groups.push({
            id: this.id,
            keys: groupKeys,
            address: this.get("address"),
            signingPrivateKey: privateGroupKey,
            adminKey: groupAdminKey,
            lastProcessedMessageId: 0,
            lastValidMessageId: 0
          });
          Snake.sessionUser.set("groups", groups);
          Snake.sessionUser.setChanged("groups");
          return Snake.sessionUser.save().result;
        }.bind(this)).then(function(){
          return this;
        }.bind(this));
      },
      insertMember: function(userId, staticPublicKey) {
        return Crypto.GroupProtocol.insertMember(userId, staticPublicKey, this);
      },
      removeMember: function(userId) {
        return Crypto.GroupProtocol.removeMember(userId, this);
      },
      promoteMemberToAdministrator: function(userId) {
        return Crypto.GroupProtocol.promoteMemberToAdministrator(userId, this);
      },
      getSessionuserGroup: function() {
        var groups = _(Snake.sessionUser.get("groups"));
        var group = groups.find(function(group) {
          return group.id === this.id;
        }, this);
        return group;
      },
      /**
       * Processes system messages for the group.
       * @memberOf Snake.Model.Group#
       * @returns {Promise} Promise resolved when all the messages have been
       *                    processed and the sessionUser's descriptor has been
       *                    updated.
       */
      processNewMessages: function() {
        Snake.Logging.debug("group", "Checking for new messages");

        // Creates the collection of messages to process
        var messagesToProcess = new Snake.Model.Collection([], {
          model: Snake.Model.Post,
          url: function() {
            return {
              object: "messages-after-id",
              addresses: [{
                address: this.id,
                id: this.getSessionuserGroup().lastProcessedMessageId
              }],
              num: 10
            };
          }.bind(this),
          comparator: Snake.dateComparator("ascending")
        });

        return messagesToProcess.fetch().result.then(function(newMessages) {
          // Removes "post" messages that do not need to be processed
          var messages = _.filter(newMessages, function(model) {
            return model.type !== "post";
          });

          // Here the order in which the messages are processesd is critical.
          // Rekeying messages must be processed in order, otherwise everything
          // blows up.
          var messagesPromise = Snake.identityPromise();
          for (var i = messages.length-1; i >= 0; i--) {
            messagesPromise = messagesPromise.then(function(i) {
              return function() {
                return this.processMessage(messages[i]);
              }.bind(this);
            }.bind(this)(i));
          }

          // If the is at least one message update the last processed message id
          // for the group and saves the sessionUser's descriptor.
          if (messages.length > 0) {
            this.getSessionuserGroup().lastProcessedMessageId =
                _(messages).max(function(message) {
              return message.id;
            }).id;

            messagesPromise.then(function() {
              Snake.sessionUser.save();
            });
          }

          return messagesPromise;
        }.bind(this));
      },
      processMessage: function(message) {
        var oldGroupKeys = this.getSessionuserGroup().keys;
        var processPromise;

        switch (message.type) {
          case "member-added":
            Snake.Logging.debug("messaging", "Handling insert rekey");
            processPromise = Crypto.GroupProtocol.
                handleInsertMessage(message.group.rekey, oldGroupKeys);
            // TODO this is not elegant
            if(message.group.rekey.rootKey) {
              this.getSessionuserGroup().lastValidMessageId = message.id;
            }
            break;

          case "member-removed":
            Snake.Logging.debug("messaging", "Handling remove rekey");
            processPromise = Crypto.GroupProtocol.
                handleRemoveMessage(message.group.rekey, oldGroupKeys);
            // TODO this is not elegant
            this.getSessionuserGroup().lastValidMessageId = message.id;
            break;

          default:
            Snake.Logging.debug("messaging", "Unknown system message received");
            Snake.Logging.debug("messaging", message);
        }

        if(processPromise) {
          return processPromise.then(function() {
            Snake.sessionUser.save();
          });
        }

        return processPromise;
      },
      /**
       * Returns true if the group is standalone, i.e. not a group of
       * friends of a user
       * @memberOf Snake.Model.Group#
       * @returns {Boolean} true, if a group of friends of a user
       */
      isStandalone: function() {
        return !this.has("friendsOf");
      },
      getSigningPrivateKey: function() {
        var privateKey, publicKey, jwkKey;

        if (!_(this.signingPrivateKey).isUndefined()) {
          privateKey = this.signingPrivateKey;
        } else {
          var descriptor = _(Snake.sessionUser.get("groups")).findByField(this.id);
          if (_(descriptor).isUndefined()
              || !_(descriptor).has("signingPrivateKey")) {
            throw new Error("You are not administrator of this group");
          }
          privateKey = descriptor.signingPrivateKey;
        }

        if (!_(this.signingPublicKey).isUndefined()) {
          publicKey = this.signingPublicKey;
        } else {
          publicKey = this.get("signingPublicKey");
        }

        jwkKey = Crypto.Common.privateKeyToJwk(privateKey, publicKey);

        return Snake.identityPromise(jwkKey);
      }
    }, {
      urlRoot: "group",
      getEncryptionKey: function(data) {
        var group = _(Snake.sessionUser.get("groups")).find(function(group) {
          return group.id === data.id;
        });
        var key = group.keys.rootKey.key;
        return Snake.identityPromise(key);
      },
      getSigningPublicKey: function(data) {
        // TODO: check against the copy the user should have in its private
        //       descriptor (but still don't)
        var rawKey = data.signingPublicKey;
        var jwkKey = Crypto.Common.ecAnsiPublicKeyToJwk(rawKey);
        return Snake.identityPromise(jwkKey);
      }
    })
  });

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for the administrative data of a group.
     * @class Snake.Model.GroupAdmin
     */
    GroupAdmin: Snake.Model.ModelBase.extend({
      initialize: function(attributes, options) {
        this.on("change:serializedBush", this.deserializeBush, this);
      },
      serializeBush: function() {
        var serializedBush = this.bush.stringify();
        this.set("serializedBush", serializedBush);

        this.bush = new Crypto.Bush();
        this.bush.parse(_.deepClone(serializedBush));
      },
      deserializeBush: function() {
        this.bush = new Crypto.Bush();
        this.bush.parse(_.deepClone(this.get("serializedBush")));
      },
      /**
       * Initializes a tree and inserts the first member into it.
       * @param {Integer} userId the id of user to insert into the tree.
       * @returns {Promise<Object>} the keys for the creator of the group.
       * @memberOf Snake.Model.GroupAdmin#
       */
      initializeBush: function(userId) {
        // Creates and initalizes the bush
        var bush = new Crypto.Bush();
        return bush.initBush().then(function() {
          // Inserts the user that has created the group
          return bush.insert(userId);
        }.bind(this)).then(function(rekeyingMessage) {
          bush.stringify();
          this.set("serializedBush", bush);

          this.bush = new Crypto.Bush();
          this.bush.parse(_.deepClone(bush));

          return rekeyingMessage.newUser;
        }.bind(this));
      },
      getSigningPrivateKey: function() {
        return Snake.identityPromise();
      }
    }, {
      urlRoot: "group-admin",
      getEncryptionKey: function(data) {
        var groups = _(Snake.sessionUser.get("groups"));
        var group = groups.find(function(group) {
          return group.id === data.id;
        });
        var key = group.adminKey;
        return Snake.identityPromise(key);
      }
    })
  });
}());
