(function() {
  "use strict";

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for a message.
     * @class Snake.Model.Message
     */
    Message: Snake.Model.ModelBase.extend({
      childModels: ["sender", "recipient"],
      idAttribute: "_fakeId",
      initialize: function(attributes, options) {
        this.options = _.defaults(options || {}, {
          fetchSender: false,
          fetchRecipient: false,
          sender: {},
          recipient: {}
        });

        this.nowAndOn(["_recipientType"], function() {
          if (this.has("_recipientType") &&
              this.get("_recipientType") !== "user") {
            throw new Error("Using message model but _recipientType is \"" +
                this.get("_recipientType") + "\"");
          }
        }.bind(this));

        if (this.options.fetchSender) {
          this.sender = new Snake.Model.User({}, this.options.sender);
          this.monitorChange("sender", "_senderId");
        }

        if (this.options.fetchRecipient) {
          this.recipient = new Snake.Model.User({}, this.options.recipient);
          this.monitorChange("recipient", "_recipientId");
        }

      },
      send: function() {
        var recipient;

        // Get the recipient/friend's descriptor
        recipient = Snake.sessionUser.getFriend(this.get("_recipientId"));

        Snake.Logging.log("messaging", "Sending message");

        return this.save().result.then(function() {
          Snake.Logging.log("messaging", "New message saved");
        }.bind(this));
      }
    }, {
      urlRoot: "message",
      handlePreSendRawData: function (data) {
        var match;

        delete data._fakeId;

        // If a message has an address manually specified it shouldn't have
        // local fields to deduce it
        if (_.has(data, "address")) {
          if (_.has(data, "_senderId")) delete data._senderId;
          if (_.has(data, "_recipientId")) delete data._recipientId;
          if (_.has(data, "_recipientType")) delete data._recipientType;
        } else if (_.has(data, "_recipientType") ||
                   _.has(data, "_recipientId")) {
          // We have to obtain the correct address given the specified local
          // fields
          if (data._recipientType === "user") {
            // User to user
            if (!_.has(data, "_senderId") ||
                data._senderId !== Snake.sessionUser.id) {
              throw new Error("Incorrect sender for message");
            }

            match = Snake.sessionUser.getFriend(data._recipientId);
            delete data._senderId;
          } else {
            throw new Error("Unknown _recipientType: " + data._recipientType);
          }

          delete data._recipientType;
          delete data._recipientId;

          if (_.isUndefined(match.otherAddress))
            throw new Error("Destination address for the user not available");

          data.address = match.otherAddress;
        }
      },
      /// It's a 1-to-1 message only if the destination address is an outgoing or
      /// incoming address associated to a friend, or if it's an initial
      /// friendship handshake message
      isMyDuty: function(urlRoot, partialData) {
        var address, match;

        if (!Snake.Model.ModelBase.isMyDuty.call(this, urlRoot))
          return false;

        address = base91.decode(partialData.address);

        if (util.memcmp(address, Snake.sessionUser.get("friendshipAddress")))
            return true;

        // User to user, current user is the recipient
        match = _(Snake.sessionUser.get("friends"))
            .findByField(address, "myAddress");

        if (!_.isUndefined(match))
          return true;

        // User to user, current user is the sender
        match = _(Snake.sessionUser.get("friends"))
            .findByField(address, "otherAddress");

        if (!_.isUndefined(match))
          return true;

        return false;
      },
      handleReceivedRawData: function(data) {
        var match, address, found = false, sessionUser;

        // TODO: move Snake.Logging.stringify in utils.js?
        // TODO: what's the whole point of _fakeId?
        data._fakeId = data.id.toString() + "-" +
          Snake.Logging.stringify(data.address);

        if (_(data).has("fhmqv") &&
            data.fhmqv.type === Snake.Crypto.FHMQV.MessageType.InitialRequest) {
          return Snake.Crypto.FHMQV.handleInitialRequest(data);
        } else if (_(data).has("address")) {
          address = data.address;

          // User to user, current user is the recipient
          match = _(Snake.sessionUser.get("friends"))
              .findByField(address, "myAddress");
          found = !_.isUndefined(match);
          if (found) {
            data._recipientType = "user";
            data._senderId = match.id;
            data._recipientId = Snake.sessionUser.id;
            return Snake.identityPromise();
          }

          // User to user, current user is the sender
          match = _(Snake.sessionUser.get("friends"))
              .findByField(address, "otherAddress");
          if (found && !_.isUndefined(match)) {
            throw new Error("Address found in multiple places");
          }
          found = !_.isUndefined(match);
          if (found) {
            data._recipientType = "user";
            data._senderId = Snake.sessionUser.id;
            data._recipientId = match.id;
            return Snake.identityPromise();
          }

          if (!found) {
            throw new Error("Received a message for an unknown address");
          }
        } else {
          throw new Error("The message doesn't have an address and it's not" +
              " a friendship request.");
        }
      },
      getEncryptionKey: function(data) {
        // TODO: this should be the only case in which we have an fhmqv field
        //       at this stage, we should probably fail here if the following
        //       conditions are not met.
        // Handle the special case of the friendship initiation request
        if (_(data).has("fhmqv") &&
            data.fhmqv.type === Snake.Crypto.FHMQV.MessageType.InitialRequest &&
            _(data.fhmqv).has("_key")) {
          return Snake.identityPromise(data.fhmqv._key);
        }

        // Message for me
        if (data._recipientId === Snake.sessionUser.id) {
          return Snake.identityPromise(
              Snake.sessionUser.getFriend(data._senderId).key);
        } else {
          // Message from me (probably)
          return Snake.identityPromise(
              Snake.sessionUser.getFriend(data._recipientId).key);
        }
      },
      getSigningPublicKey: function(data) {
        return Snake.sessionUser.getUserSigningKey(data._senderId);
      }
    })
  });

}());
