(function() {
  "use strict";

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Base class for all the system models.
     * Offers usefult features to automatically prepare requests and to
     * fetch and return attributes of child models.
     * @class Snake.Model.ModelBase
     */
    ModelBase: Backbone.Model.extend({
      idAttribute: "id",
      setId: function(newId) {
        this.set(this.idAttribute, newId);
      },
      /**
       * Loads the default values into the model.
       * @memberOf Snake.Model.ModelBase#
       */
      loadDefaultValues: function() {
        _.each(this.defaultValues || {}, function(value, key) {
          if (!this.has(key)) {
            this.set(key, _.clone(value));
          }
        }, this);
      },
      /**
       * Default implementation generating a request for a object of the
       * type specified in the urlRoot field using the current ID
       * @memberof Snake.Model.ModelBase#
       * @returns {Object} request for the server
       */
      url: function() {
        if (!this.has(this.idAttribute)) return;
        return {
          object: _.result(this, "urlRoot"),
          ids: [this.id]
        };
      },
      pendingFetches: 0,
      checkReady: function() {
        if (this.pendingFetches === 0) {
          this.trigger("ready", this);
        }
      },
      fetchAll: function () {
        var result, resolve;
        result = new Promise(function (r) { resolve = r; });
        this.once("ready", function () { resolve(); });
        this.fetch();
        return result;
      },
      // TODO: shouldn't this function invoke fetchNestedModels?
      fetch: function(options) {
        var result;

        // Wrap success function reducing the amount of pending fetches
        options = options || {};
        options.success = _(options.success).wrapAppend(function() {
          this.pendingFetches--;
          this.checkReady();
        }.bind(this));

        // Fetch myself
        this.pendingFetches++;

        result = Backbone.Model.prototype.fetch.call(this, options);

        // Snake.IO.read always returns an array, it should
        // always be a single-element so unwrap it
        return {
          result: result.result.then(function(res) {
            return res[0];
          }),
          hash: result.hash.then(function(res) {
            return res[0];
          })
        };
      },
      /**
       * Takes all the child models, as list in the childModels field, and
       * fetchs them. If a child model is a collection also fetchs all the
       * nested models of its elements.
       * @memberOf Snake.Model.ModelBase#
       */
      fetchNestedModels: function() {
        // Fetch child models with options
        if (!_.isUndefined(this.childModels)) {
          _(this.childModels).each(this.fetchNestedModel.bind(this));
        }

        this.checkReady();
      },
      fetchNestedModel: function(childModel) {
        // Check that the nested model is fetchable
        if (_.isUndefined(this[childModel]) ||
            _.isUndefined(_.result(this[childModel], "url"))) {
          return;
        }

        // When the nested model is ready remove it from the list
        this.listenTo(this[childModel], "ready", function(readyModel) {
          if (this[childModel] !== readyModel) return;
          this.stopListening(this[childModel]);
          this.pendingFetches--;
          this.checkReady();
        }.bind(this));

        // Add the nested model to be fetched to the list
        this.pendingFetches++;
        this[childModel].fetch(this.options[childModel]);
      },
      /**
       * Default implementation of the parse function.
       * Simply take the unique element if it's a single-element array
       * or just return it.
       * @memberOf Snake.Model.ModelBase#
       * @param {Object} data unparsed input object.
       * @returns {Object} the unwrapped object.
       */
      parse: function(data) {
        // If the model fetchs the data directly it will get an
        // array with a single element, otherwise if it's fetched
        // by a collection it will get right the object
        if (data instanceof Array && data.length === 1) {
          return data[0];
        } else {
          return data;
        }
      },
      /**
       * Returns a plain object containing all the model's attributes
       * plus all the attributes of the child models/collections as
       * specified in the childModels field.
       * @memberOf Snake.Model.ModelBase#
       * @returns {Object} an Object containing the model's attribute
       * along with the child models' attributes
       */
      attributeTree: function() {
        var result = _.clone(this.attributes);
        if (this.childModels) {
          _(this.childModels).each(function(childModel) {
            if (_.isUndefined(this[childModel])) return;
            if (this[childModel] instanceof Backbone.Collection) {
              result[childModel] = this[childModel].map(function(item) {
                return item.attributeTree();
              });
            } else {
              result[childModel] = this[childModel].attributeTree();
            }
          }.bind(this));
        }
        return result;
      },
      /**
       * Monitors the model for change.
       *
       * @memberof Snake.Model.ModelBase#
       * @param {String} nestedModelName name of the property containing the
       *        nested model instance, e.g. model.friends.
       * @param {String} [attribute] attribute of the current model associated to
       *        the nested model. This is the attribute which will be monitored
       *        for changes. If omitted, set to nestedModelName.
       * @param {Function} callback callback function to call in case the
       *        specfiied attribute changes.
       */
      monitorChange: function(nestedModelName, attribute, callback) {
        var nestedModelUpdated;
        attribute = attribute || nestedModelName;
        nestedModelUpdated = function() {
          if (!this.has(attribute)) return;
          if (!_(callback).isUndefined() && !callback()) return;

          // nestedModel is typically an ID or an array of ID
          var nestedModel = this.get(attribute);
          if (!_(nestedModel).isUndefined() && nestedModel.length !== 0) {

            // It's probably a single reference to another model
            // Let's set the nested model id
            if (("setId" in this[nestedModelName]) &&
                (typeof nestedModel) === "number") {
              this[nestedModelName].setId(nestedModel);
            }

            this.fetchNestedModel(nestedModelName);
          }
        }.bind(this);

        this.on("change:" + attribute, nestedModelUpdated, this);
        nestedModelUpdated();
      },
      setChanged: function(field) {
        this.trigger("change:" + field, this);
      },
      nowAndOn: function(attributes, handler) {
        var events;

        events = attributes.map(function(attribute) {
          return "change:" + attribute;
        }.bind(this)).join(" ");

        this.on(events, handler);

        if (attributes.reduce(function(prev, cur) {
          return prev || this.has(cur);
        }.bind(this), false)) {
          handler();
        }
      },
      // urlRoot should be static
      urlRoot: function() {
        return this.constructor.urlRoot;
      },
      getEncryptionKey: function() {
        if (!_(this.encryptionKey).isUndefined()) {
          return Snake.identityPromise(this.encryptionKey);
        }

        if (!_(this.constructor.getEncryptionKey).isUndefined()) {
          return this.constructor.getEncryptionKey(this.attributes);
        }

        return Snake.identityPromise();
      },
      getSigningPrivateKey: function() {
        var privateKey, publicKey, jwkKey;

        if (!_(this.signingPrivateKey).isUndefined()) {
          privateKey = this.signingPrivateKey;
        } else {
          privateKey = Snake.sessionUser.get("signingPrivateKey");
        }

        if (!_(this.signingPublicKey).isUndefined()) {
          publicKey = this.signingPublicKey;
        } else {
          publicKey = Snake.sessionUser.get("signingPublicKey");
        }

        jwkKey = Snake.Crypto.Common.privateKeyToJwk(privateKey, publicKey);

        // TODO: cache?
        return Snake.identityPromise(jwkKey);
      }
    }, {
      getSigningPublicKey: function() {
        return Snake.identityPromise();
      },
      isMyDuty: function(urlRoot, partialData) {
        return Snake.Schema.urlRootToSchemaName(urlRoot) ===
            Snake.Schema.urlRootToSchemaName(this.urlRoot);
      },
      handlePreSendRawData: function () {
        return Snake.identityPromise();
      },
      handleReceivedRawData: function () {
        return Snake.identityPromise();
      }
    })
  });
}());
