(function() {
  "use strict";

  var Crypto = Snake.Crypto;
  var IO = Snake.IO;

  Snake.Model = _.extend(Snake.Model || {}, {
    /**
     * Model for a full user, i.e. the current user
     * @class Snake.Model.SessionUser
     */
    SessionUser: Snake.Model.ModelBase.extend({
      childModels: ["friends", "posts", "groups"],
      defaultValues: {
        publicProfile: 0,
        groups: [],
        friends: [],
        profiles: []
      },
      initialize: function(attributes, options) {
        this.options = _.defaults(options || {}, {
          setDefaultValues: false,
          fetchProfiles: false,
          fetchGroups: false,
          fetchFriends: false,
          profiles: {},
          messages: {},
          groups: {},
          friends: {}
        });

        // Copies defaults values into the model
        if (this.options.setDefaultValues) {
          this.loadDefaultValues();
        }

        // Loads the list of profiles associated to this user
        if (this.options.fetchProfiles) {
          this.profiles = new Snake.Model.Collection([], {
            model: Snake.Model.Profile,
            comparator: function(profile) {
              return -profile.get("priority");
            },
            url: function() {
              if (!this.has("profiles") ||
                  this.get("profiles").length === 0) return;

              return {
                object: "profile",
                ids: _(this.get("profiles")).pluck("id")
              };
            }.bind(this)
          });

          this.monitorChange("profiles");
        }

        // Initialize the friends collection and the associated
        // event handlers
        if (this.options.fetchFriends) {
          this.friends = new Snake.Model.Collection([], {
            model: Snake.Model.User,
            url: function() {
              if (!this.has("friends") ||
                  this.get("friends").length === 0) return;
              return {
                object: "user",
                ids: _(this.get("friends")).pluck("id")
              };
            }.bind(this)
          });

          this.monitorChange("friends");
        }

        // Initialize the groups collection and the associated
        // event handlers
        if (this.options.fetchGroups) {
          this.groups = new Snake.Model.Collection([], {
            model: Snake.Model.Group,
            url: function() {
              if (!this.has("groups") || this.get("groups").length === 0)
                return;
              return {
                object: "group",
                ids: _(this.get("groups")).pluck("id")
              };
            }.bind(this)
          });

          this.monitorChange("groups");
        }

        this.nowAndOn(["signingPublicKey", "signingPrivateKey"], function() {
          if (_(this).has("signingKey")) delete this.signingKey;
        });
      },
      friendSigningKeysCache: {},
      /**
       * Fetch and, if possible, verify a user's public signing key
       * @param {Number} id id of the user whose signing key has to fetched and
       *        verified.
       * @returns {Object} an object containing:
       * * authenticated: a Boolean indicating whether we had an
       *   authenticated hash of the public key to compare
       * * key: the actual key, in the WebCryptoAPI imported format
       * @memberOf Snake.Model.SessionUser#
       */
      getUserSigningKey: function(id) {
        var isFriend, cached, user, jwkSigningKey, rawSigningKey;

        isFriend = this.isFriend(id);

        // Look for the key in the cache
        if (_(this.friendSigningKeysCache).has(id.toString())) {
          cached = this.friendSigningKeysCache[id.toString()];

          // Friendship status and authentication must match
          if (isFriend === cached.authenticated) {
            return Snake.identityPromise(cached);
          }
        }

        // Fetch the user
        user = new Snake.Model.User({id: id}, {});

        return user.fetch().result.then(function() {
          rawSigningKey = user.get("signingPublicKey");
          // Convert the key in JWK
          jwkSigningKey = Crypto.Common.ecAnsiPublicKeyToJwk(rawSigningKey);

          // Import the key
          return Crypto.Common.importKey(jwkSigningKey,
                                         "jwk",
                                         Snake.Config.Crypto.signAlgorithm,
                                         ["verify"]);
        }.bind(this)).then(function(importedSigningKey) {

          // If we're not friend we can't check, just return the key
          if (!isFriend) {
            return importedSigningKey;
          } else {
            // Hash the public key we fetch'd
            return Crypto.Common.hash(rawSigningKey).then(function(currentHash) {
              var originalHash;
              originalHash = this.getFriend(id).signingKeyHash;

              // Compare with the hash of the original public key we saw
              if (!Crypto.Common.compareHashes(currentHash, originalHash)) {
                throw new Error("The current signing public key doesn't " +
                    "match the one we authenticated.");
              }

              return importedSigningKey;
            }.bind(this));
          }
        }.bind(this)).then(function(result) {
          // Add to cache
          this.friendSigningKeysCache[id.toString()] = result;

          // Forward result
          return result;
        }.bind(this));
      },
      processMessage: function(message, hash) {
        var answer, friendship, protocolPromises, senderId, groups;

        senderId = message.get("_senderId");
        groups = Snake.sessionUser.get("groups");

        if (this.isFriend(senderId)) {
          friendship = this.getFriend(senderId);
          // Update the ID of the last processed message for this friend

          // TODO: isn't it an error if this condition does not hold?
          if (message.get("id") > friendship.lastProcessedMessageId) {
            friendship.lastProcessedMessageId = message.get("id");
          } else {
            throw new Error("We got a message older than the last registered " +
                "one: got " + message.id.toString() + ", while the last seen " +
                "was " + friendship.lastProcessedMessageId.toString());
          }
        }

        // TODO: switch to enum?
        switch (message.get("type")) {
          case "post":
            if (_.isUndefined(friendship)) {
              throw new Error("Message from unknown friend");
            }

            return message;

          case "friendship-message":

            if (util.memcmp(message.get("address"),
                Snake.sessionUser.get("friendshipAddress")) &&
                message.get("id") >
                Snake.sessionUser.get("lastFriendshipAddressIndex")) {
              Snake.sessionUser.set("lastFriendshipAddressIndex",
                message.get("id"));
              // TODO: should we save here?
            }

            // Prepare the answer
            answer = new Snake.Model.Message({
              _recipientType: "user",
              _recipientId: senderId,
              _senderId: Snake.sessionUser.id,
              type: "friendship-message"
            }, {});

            return Snake.identityPromise().then(function () {
              if (_.isUndefined(friendship))
                return this.newFriendship(senderId, message.get("id"));
            }.bind(this)).then(function (newFriendship) {
              if (_.isUndefined(friendship)) {
                friendship = newFriendship;

                answer.set("newAddress", friendship.myAddress);
                // TODO: can we avoid this check here with a proper validation?
                if (!message.has("newAddress")) {
                  throw new Error("A message from an unknown user doesn't " +
                      "contain an answer address");
                }
              }

              // TODO: keep track of old addresses!
              if (message.has("newAddress")) {
                friendship.otherAddress = message.get("newAddress");
              }
              answer.set("address", friendship.otherAddress);

              // Update the cumulative hash using the hash
              // of the message just received
              return this.updateCumulativeHash(friendship.handshakeHash, hash);
            }.bind(this)).then(function(newHash) {
                friendship.handshakeHash = newHash;

                // Delegate processing to Snake.Crypto
                // Order of the protocols used to matter, but now it shouldn't
                protocolPromises = Crypto.FriendshipProtocol.handleProtocols(
                    [Crypto.FHMQV, Crypto.SMP], message, friendship,
                    senderId);

                // FHMQV and SMP are carried out in parallel, their results
                // (i.e. what they have to save locally and what they have to
                // send to the other party) are managed by applyResults
                return this.applyResults(protocolPromises,
                    Snake.Model.SessionUser.HandshakeMessageParts,
                    answer, friendship);
            }.bind(this));
          case "group-joined":
            var newGroup = message.get("group");

            // Adds the information of the group into the sessionUser's groups
            Snake.sessionUser.get("groups").push({
              id: newGroup.groupId,
              keys: newGroup.keys,
              address: newGroup.address,
              lastProcessedMessageId: 0,
              lastValidMessageId: newGroup.lastValidMessageId
            });

            // TODO raw ids are not for an human user
            // TODO allow the user to refuse this?
            // Notifies to the user that he has been added to the group
            Snake.Modal.enqueue("Group joined",
                "You have been added to the group " +
                newGroup.groupId + " by " + newGroup.addedBy);

            Snake.sessionUser.setChanged("groups");
            return Snake.sessionUser.save().result;

          // TODO: s/leaved/left/
          case "group-leaved":
            // TODO: we should check the sender is actually allowed to kick us
            //       out (and invite us?)
            var removedGroup = message.get("group");

            // Removes the group from the sessionUser's groups
            groups.splice(_.indexOf(groups, _.find(groups, function(group) {
              return group.id === removedGroup.groupId;
            })), 1);
            Snake.sessionUser.setChanged("groups");
            Snake.sessionUser.save();

            // TODO raw ids are not for an human user
            // Notifies to the user that he has been removed from the group
            Snake.Modal.enqueue("Group leaved",
                "You have been remove form the group " +
                removedGroup.groupId + " by " + removedGroup.removedBy);

            break;

          case "group-admin-access":
            var adminGroup = message.get("group");

            // Saves the administrative keys for the group
            _.extend(_.findByField(groups, adminGroup.groupId, "id"), {
              signingPrivateKey: adminGroup.signingPrivateKey,
              adminKey: adminGroup.adminKey
            });

            // TODO raw ids are not for an human user
            // Notifies to the user the he has been promoted as the administrator
            // of a group
            Snake.Modal.enqueue("Group admin access",
                "You have received the administrative power for the group" +
                adminGroup.groupId + " by " + adminGroup.promotedBy);
            break;

          default:
            Snake.Logging.debug("messaging",
                "Unknown system message received");
            Snake.Logging.debug("messaging", message);
        }
      },
      // TODO: make this return a promise that everything will finish
      // TODO: put a lock here, we don't want to check messages concurrently or
      //       bad things will happen
      processNewMessages: function() {
        var messagesToProcess, messagesResult, processedResult;

        Snake.Logging.debug("messaging", "Checking for new messages");

        // TODO: set up friendshipAddress recycling
        // Query to select the messages we haven't processed yet
        messagesToProcess = new Snake.Model.Collection([], {
          model: Snake.Model.Message,
          url: function() {
            var fromFriends, friendshipRequests;

            if (!this.has("id")) return;

            fromFriends = _.map(this.get("friends"), function (friend) {
              return {
                address: friend.myAddress,
                id: friend.lastProcessedMessageId
              };
            }.bind(this));

            friendshipRequests = [{
                address: Snake.sessionUser.get("friendshipAddress"),
                id: Snake.sessionUser.get("lastFriendshipAddressIndex")
            }];

            return {
              object: "messages-after-id",
              addresses: fromFriends.concat(friendshipRequests),
              num: 10
            };
          }.bind(this),
        });

        // Fetch new messages (and compute their hash)
        messagesResult = messagesToProcess.fetch({hash: true});

        // Wait for the messages and their hashes
        return Promise.every(messagesResult.result, messagesResult.hash)
            .then(function(promisesResults) {
          var promises, toProcess;

          toProcess = _.sortBy(_.map(messagesToProcess.models,
                                     function (a, i) { return [i, a]; }),
                               function (a) { return a[1].get("id"); });

          // Process each message and get back a promise
          promises = _(toProcess).map(function(val) {
            var i = val[0];
            var message = val[1];

            return this.processMessage(message, promisesResults[1][i]);
          }.bind(this));

          // Filter out messages parsed on the fly (no promise)
          promises = _(promises).filterUndefined();

          if (promises.length > 0) {
            return Promise.every.apply(Promise, promises);
          } else {
            return Snake.identityPromise([]);
          }
        }.bind(this)).then(function(result) {
          processedResult = result;
          if (messagesToProcess.length > 0) {
            // TODO: ensure that we always need to update friends list
            this.setChanged("friends");
            return this.save().result;
          }
        }.bind(this)).then(function() {
          return _(processedResult).filterUndefined();
        }.bind(this)).catch(Snake.errorHandler.bind("messaging"));

      },
      getGroup: function(id) {
        var result;
        result = _(this.get("groups")).findByField(id);

        if (_(result).isUndefined()) {
          throw new Error("Can't find group");
        }

        return result;
      },
      getFriend: function(id) {
        var result;
        result = _(this.get("friends")).findByField(id);

        if (_(result).isUndefined()) {
          throw new Error("Can't find friendship");
        }

        return result;
      },
      isFriend: function(id) {
        var result;
        result = _(this.get("friends")).findByField(id);

        return !(_(result).isUndefined());
      },
      // Note: does NOT call setChanged("friends")
      newFriendship: function(id, firstMessageId) {
        var myAddress, authenticate, newFriend;

        if (this.isFriend(id)) {
          throw new Error("Friendship already exists.");
        }

        myAddress = Snake.generateAddress();

        // We are getting to know a new user, let's check if we can validate its
        // public key offline (i.e. non-interactively)
        authenticate = function (authenticator) {
          return function() {
            if (authenticator.isEnabled()) {
              return authenticator.verify(id).then(function (res) {
                return res === Crypto.OfflineAuthenticator.AuthResult.Valid;
              });
            }
            return false;
          };
        };

        // We need to fetch the user public key to populate the Web of Trust
        newFriend = new Snake.Model.User({id: id}, {});

        return Snake.firstPromise(_.map(
          [Crypto.WebOfTrust, Crypto.ManualAuthenticator], authenticate))
          .then(function (successfulPromise) {
            return newFriend.fetch().result.then(function() {
              return {
                "id": id,
                "myAddress": myAddress,
                "lastProcessedMessageId": firstMessageId,
                "friendshipProgress": Snake.Model.SessionUser.FriendshipProgress.None,
                "isAuthenticated": successfulPromise,
                "staticPublicKey": newFriend.get("staticPublicKey")
              };
            });
          });
      },
      becomeFriend: function(id) {
        var message, other, myAddress;

        if (this.isFriend(id)) {
          throw new Error("You are already friends.");
        }

        myAddress = Snake.generateAddress();
        other = new Snake.Model.User({id: id}, {});

        // Fetch other user's data
        return other.fetch().result.then(function () {
          return this.newFriendship(id, -1);
        }.bind(this)).then(function (friendship) {
          // Once we got the friendship object, we can compose the message and
          // start the friendship protocols

          // TODO: check correctness of friendshipAddress?
          message = new Snake.Model.Message({
            address: other.get("friendshipAddress"),
            _recipientType: "user",
            _recipientId: id,
            newAddress: friendship.myAddress,
            type: "friendship-message"
          }, {});

          // Wait for FHMQV-C and SMP to complete
          return this.applyResults(
              [Crypto.FHMQV.initiate(id),
               Crypto.SMP.step1(id, friendship.isAuthenticated)],
              Snake.Model.SessionUser.HandshakeMessageParts,
              message, friendship);
        }.bind(this)).then(function(result) {
          this.setChanged("friends");
          return this.save().result.then(function(){
            return result;
          });
        }.bind(this)).catch(Snake.errorHandler.bind("messaging"));
      },
      /**
       * Manage results from a FriendshipProtocol (i.e. FHMQV or SMP).
       *
       * @param {Promise[]} promises promises of the various
       *        FriendshipProtocols. Will resolve when they have performed the
       *        necessary operations.
       * @param {String[]} names contains the list of names of the
       *        corresponding entries in the promises array.
       * @param {Snake.Model.Message} message the message to populate with the
       *        information collected from the FriendshipProtocols.
       * @param {Friend} recipient reference to the private descriptor of the
       *        friend we're dealing with.
       * @returns {Promise} a Promise which will complete when all the data
       *          generated by the FriendshipProtocol handlers has been
       *          sent/stored.
       * @memberOf Snake.Model.SessionUser#
       */
      applyResults: function(promises, names, message, recipient) {
        var promiseResults, recipientId, temporaryKey, groupModels;

        return Promise.every.apply(this, promises).then(function(res) {
          var sentResult, somethingToSend = false;
          promiseResults = res;

          // Expand message with results from promises
          res.forEach(function(promiseResult, i) {
            if (!_(promiseResult).isUndefined()) {
              Snake.Logging.log("friendship",
                  names[i] + ".store: " +
                  Snake.Logging.stringify(promiseResult.store) + "\n" +
                  names[i] + ".send: " +
                  Snake.Logging.stringify(promiseResult.send));
            }

            if (_(promiseResult).has("store") &&
                _(promiseResult.store).has("common") &&
                _(promiseResult.store.common).has("key")) {
              temporaryKey = promiseResult.store.common.key;
            }

            if (!_(promiseResult).isUndefined() &&
                _(promiseResult).has("send") &&
                promiseResult.send !== null) {
              if ("common" in promiseResult.send) {
                _(_(promiseResult.send.common).keys()).each(function(key) {
                  message.set(key, promiseResult.send.common[key]);
                });
                delete promiseResult.send.common;
              }

              message.set(names[i], promiseResult.send);
              somethingToSend = true;
            }
          });

          // If we have something to send...
          if (somethingToSend) {
            // At this point we're sure we have the recipientId
            recipientId = message.get("_recipientId");
            recipient.id = recipientId;

            if (!_(temporaryKey).isUndefined()) {
              message.encryptionKey = temporaryKey;
            }

            // Send & update cumulative hash
            sentResult = message.save(null, {hash: true});
            return Promise.every(sentResult.result, sentResult.hash).then(
                function(results) {
                  return this.updateCumulativeHash(recipient.handshakeHash,
                      results[1]);
                }.bind(this)).then(function(newHandshakeHash) {
              recipient.handshakeHash = newHandshakeHash;
            });
          }
        }.bind(this)).then(function() {
          var body, profileData;

          // Ask what profiles the user want to give him
          if (!_(recipient).has("shownProfiles")) {
            // Public profiles are always available
            profileData = _(Snake.sessionUser.profiles.models).filter(function(
                profile) {
              return profile.get("isPublic") !== true;
            });

            // Collect information to show
            profileData = _(profileData).map(function(profile) {
              return {
                id: profile.id,
                name: profile.get("name"),
                surname: profile.get("surname")
              };
            });

            return Snake.Model.SessionUser.chooseProfiles(profileData)
                .then(function (profilesToShow) {
              recipient.shownProfiles = profilesToShow;
            });
          }
        }.bind(this)).then(function() {
          var done = true;

          // Override friendship data with the results from promises
          promiseResults.forEach(function(promiseResult, i) {

            if (_(promiseResult).isUndefined() ||
                !_(promiseResult).has("store")) return;

            done = done && ("done" in promiseResult);

            if (promiseResult.store !== null
                && "common" in promiseResult.store) {
              _(recipient).extend(promiseResult.store.common);
              delete promiseResult.store.common;
            }

            if (promiseResult.store === null
                || Object.keys(promiseResult.store).length === 0) {
              if (names[i] in recipient) delete recipient[names[i]];
            } else {
              recipient[names[i]] = promiseResult.store;
            }
          });

          // Check if all the protocols did their job
          if (done) {
            Snake.Logging.log("friendship", "We have a new friend!");
            // Update friendshipProgress
            recipient.friendshipProgress =
                Snake.Model.SessionUser.FriendshipProgress.Complete;

            // Drop all the storages for the protocol runs
            _(names).each(function(field) {
              if (!_(recipient[field]).isUndefined()) {
                delete recipient[field];
              }
            });

            // Finalize the key
            return this.updateCumulativeHash(recipient.handshakeHash,
                recipient.key).then(function(finalKey) {
              recipient.key = finalKey;
              delete recipient.handshakeHash;

              this.setFriendship(recipient);
            }.bind(this)).then(function() {
              var groups;
              // Take all the groups of all the shown profiles and merge them
              groups = _(_(Snake.sessionUser.profiles.models)
                  .map(function(profile) {
                    if (recipient.shownProfiles.indexOf(profile.id) !== -1) {
                      return _(profile.get("groups")).pluck("id");
                    } else {
                      return [];
                    }
                  })).flatten();

              // Fetch all groups from the array
              groupModels = new Snake.Model.Collection([], {
                model: Snake.Model.Group,
                url: function() {
                  return {
                    object: "group",
                    ids: groups
                  };
                }.bind(this),
                fetchAdminData: true
              });

              return groupModels.fetch().result;
            }.bind(this)).then(function() {
              return Promise.all(groupModels.map(function(groupModel){
                // HACK because adminData doesn't get fetched on model init
                // see group.js.initialize
                return groupModel.adminData.fetch().result;
              }));
            }).then(function() {
              return Promise.all(groupModels.map(function(group){
                return group.insertMember(recipient.id,
                                          recipient.staticPublicKey);
              }));
            });
          } else {
            this.setFriendship(recipient);
          }
        }.bind(this));
      },
      updateCumulativeHash: function(oldHash, newHash) {
        var result;

        // Update the cumulative hashing of the handshake
        if (_(oldHash).isUndefined()) {
          // Just for debugging purposes
          oldHash = new Uint8Array(0);
          result = Snake.identityPromise(newHash);
        } else {
          // Compute the MAC of the new hash using the old hash as key
          result = Crypto.Common.mac(oldHash, newHash);
        }

        result.then(function(newCumulativeHash) {
          Snake.Logging.log("friendship",
              "Old cumulative hash: " + util.abv2hex(oldHash) + "\n" +
              "New message hash: " + util.abv2hex(newHash) + "\n" +
              "New cumulative hash: " + util.abv2hex(newCumulativeHash));
        });

        return result;
      },
      setFriendship: function(friendship) {
        var property, friend, friends;

        // Delete existing properties with value "undefined"
        for (property in friendship) {
          if (friendship.hasOwnProperty(property) &&
              _(friendship[property]).isUndefined()) {
            delete friendship[property];
          }
        }

        // Ensure update
        friends = this.get("friends");
        if (this.isFriend(friendship.id)) {
          friend = this.getFriend(friendship.id);
          friends.splice(friends.indexOf(friend), 1);
        }
        friends.push(friendship);

        return friendship;
      },
      unfriend: function(id) {
        var friends, friend;

        if (!this.isFriend(id)) {
          throw new Error("You're not friends.");
        }

        friend = this.getFriend(id);
        friends = this.get("friends");
        friends.splice(friends.indexOf(friend), 1);

        // TODO: setChanged should be replaced by getClone
        this.setChanged("friends");
        this.save();
      },
      /**
       * Returns the data that need to be authenticated to verify the identity
       * of the user.
       * @returns {ArrayBufferView} the requested data.
       * @memberOf Snake.Model.SessionUser#
       */
      // TODO: make me private and fix those who use me switching to
      // getAuthenticationToken
      getDataToAuthenticate: function() {
        var id;
        // TODO: include all static public fields?
        id = this.id.toString(16);
        id = ((id.length % 2 !== 0) ? "0" : "") + id;
        id = util.hex2abv(id);

        return Snake.concatBuffers(this.get("staticPublicKey"), id);
      },
      getAuthenticationToken: function() {
        return Crypto.Common.hash(this.getDataToAuthenticate());
      }
    },
    // Static properties
    {
      FriendshipProgress: {
        "None": 0,
        "HandshakeSent": 1,
        "HandshakeAnswered": 2,
        "Complete": 3
      },
      TrustLevel: {
        "None": 0
      },
      HandshakeMessageParts: ["fhmqv", "smp"],
      urlRoot: "session-user",
      isMyDuty: function(urlRoot, partialData) {
        return Snake.Model.ModelBase.isMyDuty.call(this, urlRoot);
      },
      getSigningPublicKey: function(data) {
        var rawKey, jwkKey;

        if (!_.isUndefined(Snake.sessionUser.signingPublicKey)) {
          rawKey = Snake.sessionUser.signingPublicKey;
        } else {
          rawKey = Snake.sessionUser.get("signingPublicKey");
        }

        jwkKey = Crypto.Common.ecAnsiPublicKeyToJwk(rawKey);

        return Snake.identityPromise(jwkKey);
      },
      getEncryptionKey: function(data) {
        if (_(Snake.sessionUser.encryptionKey).isUndefined()) {
          throw new Error("No key available!");
        }

        return Snake.identityPromise(Snake.sessionUser.encryptionKey);
      },
      register: function(username, password) {
        var newUser, keyPairsPromise, keyPromise, deleteKeyPromise, defaultGroup,
            privatePublicProfileDescriptor, defaultProfileKeyPromise,
            privateDefaultProfileDescriptor, publicProfile, defaultProfile;

        // Initialize the new user and the private descriptors
        newUser = new Snake.Model.SessionUser({
          username: username,
          friendshipAddress: Snake.generateAddress(),
          lastFriendshipAddressIndex: -1
        }, {
          setDefaultValues: true
        });
        privatePublicProfileDescriptor = {};
        privateDefaultProfileDescriptor = {};

        // Set Snake.sessionUser to newUser
        Snake.sessionUser = newUser;

        // Create an empty default group
        defaultGroup = new Snake.Model.Group({}, {setDefaultValues: true});

        // Initialize profiles
        publicProfile = new Snake.Model.Profile({
          name: username,
          surname: "public"
        }, {setDefaultValues: true});
        defaultProfile = new Snake.Model.Profile({
          name: username,
          surname: "default"
        }, {setDefaultValues: true});

        // Generate the necessary asymmetric keys for the user descriptor
        keyPairsPromise = Promise.every(
            Crypto.Common.generateSigningKeyPair(),
            Crypto.Common.generateKeyAgreementKeyPair(),
            Crypto.Common.generateKeyAgreementKeyPair())
            .then(function(keyPairs) {
          // Save the three generated asymmetric keys
          _(["static", "ephemeral", "signing"]).each(function(field, i) {
            newUser.set(field + "PublicKey",
                Crypto.Common.jwkToAnsiPublicKey(keyPairs[i].publicKey));
            newUser.set(field + "PrivateKey",
                Crypto.Common.jwkToPrivateKey(keyPairs[i].privateKey));
          });
        });

        // Generate salt and derive key from password
        keyPromise = Crypto.Common.generatePasswordSalt()
            .then(function(salt) {
          newUser.set("salt", salt);
          return Crypto.Common.deriveKeyFromPassword(util.utf8encode(password),
                                                     salt);
        }).then(function(key) {
          newUser.encryptionKey = key;
        });

        // Generates key for object deletions
        deleteKeyPromise = Crypto.Common.generateHMACKey()
            .then(function(removeKey) {
          newUser.set("deleteKey", removeKey);
        });

        // TODO: public profile should not be encrypted
        // Generate profiles keys
        defaultProfileKeyPromise = Crypto.Common.generateSymmAuthKey()
            .then(function(key) {
          privateDefaultProfileDescriptor.key = key;
          defaultProfile.encryptionKey = key;
        });

        return Promise.every(keyPromise, keyPairsPromise, deleteKeyPromise)
            .then(function() {
          // Save the new user
          return newUser.save({}, {rawRequest: true}).result;
        }).then(function() {
          var groupData;
          // We have the user ID
          defaultProfile.set("userId", newUser.id);
          publicProfile.set({"userId": newUser.id, isPublic: true});

          // OK at this point we should setup the default group
          // TODO: set name
          groupData = {name: "Asdomare's friends", description: ""};
          return defaultGroup.setupNew(groupData);
        }).then(function() {
          defaultGroup.set("friendsOf", newUser.id);
          defaultProfile.set("groups", [{id: defaultGroup.id}]);
          return defaultGroup.save().result;
        }).then(function(){
          // Create the public and default profiles
          return defaultProfileKeyPromise;
        }).then(function() {
          // Save them
          return Promise.every(publicProfile.save().result,
              defaultProfile.save().result);
        }).then(function() {
          // We now have the ID of the two profiles
          privatePublicProfileDescriptor.id = publicProfile.id;
          privateDefaultProfileDescriptor.id = defaultProfile.id;
          newUser.set("publicProfile", publicProfile.id);

          // Private profiles
          newUser.set("profiles", [privatePublicProfileDescriptor,
            privateDefaultProfileDescriptor]);

          newUser.set("friendshipAddress", Snake.generateAddress());

          // Re-save the user
          return newUser.save().result;
        }.bind(this));
      },
      login: function(username, password) {
        var response, me;

        return Snake.identityPromise().then(function () {
          return IO.read({url: {
            object: "username",
            username: username
          }}, {rawResponse: true}).result;
        }.bind(this)).then(function(rawResponse) {
          response = _(rawResponse[0]).pick(Snake.Schema.user.publicFields);

          // Sets the session user and removes the login view
          Snake.sessionUser = new Snake.Model.SessionUser(response, {
            fetchProfiles: true,
            fetchGroups: true,
            profiles: {
              fetchFriendsGroup: true
            }
          });

          me = Snake.sessionUser;

          return Snake.Crypto.Common.deriveKeyFromPassword(
                util.utf8encode(password), IO.deserializeBinary(response.salt));

        }.bind(this)).then(function(key) {
          me.encryptionKey = key;
          me.signingPublicKey = IO.deserializeBinary(response.signingPublicKey);

          return me.fetch().result;
        }.bind(this)).then(function (userData) {
          var result, accept;

          // Fetch has failed, most probably the password was wrong
          if (_.isUndefined(userData)) {
            throw new Error("Couldn't load user info. " +
                            "Check username and password");
          }

          result = new Promise(function (resolve, reject) {
            accept = resolve;
          });

          // TODO: we're not using reject, is something goes wrong beyond this
          //       point we won't notice it
          me.listenToOnce(Snake.sessionUser, "ready", function() {
            accept(me);
          }, this);

          return result;
        }.bind(this));
      }
    })
  });
}());
