(function() {
  "use strict";

  /**
   *
   * @param {type} schema
   * @returns {}
   */
  function extendBaseSchema(schema) {

    schema["$schema"] = "http://json-schema.org/draft-04/schema#";
    schema["type"] = "object";

    schema.properties = schema.properties || {};
    schema.properties["id"] = {type: "integer"};
    schema.properties["date"] = {type: "string"};

    schema.required = schema.required || [];
    schema.required.push("id");
    schema.required.push("date");

    schema.publicFields = schema.publicFields || [];
    schema.publicFields.push("object", "id", "date");

    return schema;
  };

  function localNotRequired(item) {
    _(item).forEach(function (value, key) {
      var i;
      if (key === "required" && _(value).isArray()) {
        i = value.length;
        while (i--) {
          if (_.isString(value[i]) && value[i][0] === "_") {
            value.splice(i, 1);
          }
        }
      } else if (_.isObject(value)) {
        localNotRequired(value);
      }
    });
  }

  /**
   * Namespace for the schema describing the Snake object model.
   * @namespace Snake.Schema
   */
  Snake.Schema = Snake.Schema || {};

  // Indentation abuse for the sake of readibility, sorry for that
  Snake.Schema.init = function() {
    tv4.addFormat({
      "public-key": function(data, schema) {
        // TODO: should we check the length with Snake.Config?
        if (data instanceof Uint8Array && data.length % 2 === 1
            && data[0] === 4) return null;

        return "Only ANSI key in Uint8Array are supported.";
      },
      "binary": function(data, schema) {
        return data instanceof Uint8Array ? null :
            "Private keys must be an Uint8Array";
      },
      "address": function (data, schema) {
        // TODO: validate appropriately
        return null;
      }
    });

    Snake.Schema = _.extend(Snake.Schema, {
      binaryFormats: ["binary", "public-key", "address"],
      // User descriptor
      // ===============
      user: extendBaseSchema({
        properties: {
          username: {type: "string", pattern: "^[a-zA-Z0-9]{3,}$"},
          salt: {type: "object", format: "binary"},
          staticPublicKey: {type: "object", format: "public-key"},
          ephemeralPublicKey: {type: "object", format: "public-key"},
          signingPublicKey: {type: "object", format: "public-key"},
          staticPrivateKey: {type: "object", format: "binary"},
          ephemeralPrivateKey: {type: "object", format: "binary"},
          signingPrivateKey: {type: "object", format: "binary"},
          deleteKey: {type: "object", format: "binary"},
          publicProfile: {type: "integer"},
          friendshipAddress: {type: "object", format: "address"},
          lastFriendshipAddressIndex: {type: "integer"},
          groups: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/group"}
          },
          friends: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/friend"}
          },
          profiles: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/profile"}
          }
        },
        required: [
          "username",
          "staticPublicKey",
          "ephemeralPublicKey",
          "signingPublicKey",
          "publicProfile"
        ],
        publicFields: [
          "salt", "username", "staticPublicKey", "ephemeralPublicKey",
          "signingPublicKey", "publicProfile", "friendshipAddress"
        ],
        authenticationOnlyFields: [],
        additionalProperties: false,
        definitions: {
          bushKey: {
            type: "object",
            properties: {
              key: {type: "object", format: "binary"},
              id: {type: "integer"}
            },
            additionalProperties: false,
            required: ["id", "key"]
          },
          groupKeys: {
            type: "object",
            properties: {
              rootKey: {type: "object", $ref: "#/definitions/bushKey"},
              auxiliaryKeys: {
                type: "array",
                items: {type: "object", $ref: "#/definitions/bushKey"}
              }
            }
          },
          // TODO: recheck all "required"
          group: {
            type: "object",
            properties: {
              id: {type: "integer"},
              lastProcessedMessageId: {type: "integer"},
              lastValidMessageId: {type: "integer"},
              keys: {type: "object", $ref: "#/definitions/groupKeys"},
              signingPrivateKey: {type: "object", format: "binary"},
              adminKey: {type: "object", format: "binary"},
              address: {type: "object", format: "address"}
            },
            additionalProperties: false,
            required: [
              "id", "keys", "lastProcessedMessageId", "lastValidMessageId",
              "address"
            ]
          },
          friend: {
            type: "object",
            properties: {
              id: {type: "integer"},
              staticPublicKey: {type: "object", format: "public-key"},
              key: {type: "object", format: "binary"},
              signingKeyHash: {type: "object", format: "binary"},
              handshakeHash: {type: "object", format: "binary"},
              // TODO: lastReadMessageId
              lastProcessedMessageId: {type: "integer"},
              friendshipProgress: {type: "integer"},
              otherAddress: {type: "object", format: "address"},
              myAddress: {type: "object", format: "address"},
              isAuthenticated: {type: "boolean"},
              shownProfiles: {
                type: "array",
                items: {type: "integer"}
              },
              fhmqv: {
                type: "object",
                properties: {
                  expectedConfirmationToken: {
                    type: "object",
                    format: "binary"
                  },
                  key: {type: "object", format: "binary"},
                  expectation: {type: "integer"}
                },
                additionalProperties: false
              },
              smp: {
                type: "object",
                properties: {
                  expectation: {type: "integer"},
                  privateKey2: {type: "object", format: "binary"},
                  privateKey3: {type: "object", format: "binary"},
                  aToB: {type: "object", $ref: "#/definitions/smp"},
                  bToA: {type: "object", $ref: "#/definitions/smp"}
                },
                required: ["expectation"],
                additionalProperties: false
              }
            },
            additionalProperties: false,
            // TODO: more required fields here
            required: ["id"]
          },
          smp: {
            type: "object",
            properties: {
              // TODO: maybe someone is a point on an EC
              sharedSecret: {type: "object", format: "binary"},
              p: {type: "object", format: "binary"},
              otherP: {type: "object", format: "binary"},
              q: {type: "object", format: "binary"}
            },
            additionalProperties: false
          },
          profile: {
            type: "object",
            properties: {
              // TODO add other properties
              id: {type: "integer"},
              key: {type: "object", format: "binary"}
            },
            additionalProperties: false,
            required: ["id"]
          }
        }
      }),
      // User profile
      // ============
      profile: extendBaseSchema({
        properties: {
          userId: {type: "integer"},
          name: {
            type: "string",
            pattern: "^.+$",
            title: "Name",
            description: "your name",
            input: "text"
          },
          surname: {
            type: "string",
            pattern: "^.+$",
            title: "Surname",
            description: "your surname",
            input: "text"
          },
          email: {
            type: "string",
            title: "Email",
            description: "your email",
            input: "email"
          },
          birthday: {
            type: "string",
            title: "Birthday",
            description: "your birthday date",
            input: "date"
          },
          priority: {
            type: "integer"
          },
          groups: {
            type: "array",
            items: {
              type: "object",
              $ref: "#/definitions/group"
            }
          },
          isPublic: {type: "boolean"} // FIXME: this is ugly
        },
        publicFields: [],
        authenticationOnlyFields: ["isPublic"],
        required: ["name", "surname", "birthday", "priority", "groups", "userId"],
        formFields: ["name", "surname", "email", "birthday"],
        additionalProperties: false,
        definitions: {
          group: {
            type: "object",
            properties: {
              // TODO add other properties
              id: {type: "integer"},
              additionalProperties: false,
              required: ["id"]
            }
          }
        }
      }),
      // Group descriptor
      // ================
      group: extendBaseSchema({
        properties: {
          signingPublicKey: {
            type: "object",
            format: "public-key"
          },
          name: {
            type: "string",
            title: "Name",
            description: "name of the group",
            input: "text"
          },
          description: {
            type: "string",
            title: "Description",
            description: "group description",
            input: "textarea"
          },
          friendsOf: {type: "integer"},
          members: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/member"}
          },
          administrators: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/member"}
          },
          oldKeys: {
            type: "array",
            items: {type: "object", $ref: "#/definitions/key"}
          },
          address: {type: "object", format: "address"}
        },
        required: ["name", "members", "oldKeys", "address"],
        formFields: ["name", "description"],
        publicFields: ["signingPublicKey"],
        authenticationOnlyFields: [],
        additionalProperties: false,
        definitions: {
          member: {
            type: "object",
            properties: {
              id: {type: "integer"},
              staticPublicKey: {type: "object", format: "public-key"},
              additionalProperties: false,
              required: ["id", "staticPublicKey"]
            }
          },
          key: {
            type: "object",
            properties: {
              date: {type: "string"},
              key: {type: "object", format: "binary"}
            }
          }
        }
      }),
      // Group administrator descriptor
      // ==============================
      "group-admin": extendBaseSchema({
        properties: {
          // TODO: TODO
          serializedBush: {type: "object"}
        },
        required: ["serializedBush"],
        publicFields: [],
        authenticationOnlyFields: [],
        additionalProperties: false
      }),
      // WebOfTrustEntry descriptor
      // ================
      "wot-entry": extendBaseSchema({
        properties: {
          userIdMac: {type: "object", format: "binary"},
          groupIdMac: {type: "object", format: "binary"},
          editToken: {type: "object", format: "binary"}
        },
        required: [],
        publicFields: ["userIdMac", "groupIdMac", "editToken"],
        authenticationOnlyFields: [],
        additionalProperties: false,
      }),
      // Message
      // =======
      // TODO: enforce type with oneOf
      message: extendBaseSchema({
        properties: {
          address: {type: "object", format: "address"},
          _recipientType: {enum: ["user", "group"]},
          _recipientId: {type: "integer"},
          _senderId: {type: "integer"},
          type: {
            enum: [
              "post", "friendship-message", "group-joined", "group-leaved",
              "member-added", "member-removed", "group-admin-access"
            ]
          },
          // TODO: s:post:body:
          post: {type: "string"},
          authorId: {type: "integer"},
          newAddress: {type: "object", format: "address"},
          fhmqv: {
            type: "object",
            oneOf: [
              // InitialRequest message
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.FHMQV.MessageType.InitialRequest,
                    minimum: Snake.Crypto.FHMQV.MessageType.InitialRequest
                  },
                  ephemeralPublicKey: {type: "object", format: "public-key"},
                  confirmationToken: {type: "object", format: "binary"},
                  identity: {type: "object", format: "binary"},
                  _confirmationToken: {type: "object", format: "binary"},
                  _key: {type: "object", format: "binary"},
                  _signingKeyHash: {type: "object", format: "binary"}
                },
                additionalProperties: false,
                required: [
                  "ephemeralPublicKey",
                  "confirmationToken",
                  "identity",
                  "type",
                  "_confirmationToken",
                  "_key",
                  "_signingKeyHash"
                ]
              },
              // ConfirmationToken message
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.FHMQV.MessageType.ConfirmationToken,
                    minimum: Snake.Crypto.FHMQV.MessageType.ConfirmationToken
                  },
                  confirmationToken: {type: "object", format: "binary"}
                },
                additionalProperties: false,
                required: ["confirmationToken", "type"]
              }
            ]
          },
          smp: {
            type: "object",
            oneOf: [
              // Message 1
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.SMP.MessageType.Step1,
                    minimum: Snake.Crypto.SMP.MessageType.Step1
                  },
                  publicKey2: {type: "object", format: "public-key"},
                  publicKey3: {type: "object", format: "public-key"},
                  aToB: {
                    type: "object",
                    properties: {
                      question: {type: "string"}
                    },
                    additionalProperties: false,
                    required: ["question"]
                  }
                },
                additionalProperties: false,
                required: ["type", "publicKey2", "publicKey3"]
              },
              // Message 2
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.SMP.MessageType.Step2,
                    minimum: Snake.Crypto.SMP.MessageType.Step2
                  },
                  publicKey2: {type: "object", format: "public-key"},
                  publicKey3: {type: "object", format: "public-key"},
                  aToB: {
                    type: "object",
                    properties: {
                      p: {type: "object", format: "binary"},
                      q: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["p", "q"]
                  },
                  bToA: {
                    type: "object",
                    properties: {
                      question: {type: "string"},
                      p: {type: "object", format: "binary"},
                      q: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["p", "q"]
                  }
                },
                additionalProperties: false,
                required: ["type", "publicKey2", "publicKey3"]
              },
              // Message 3
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.SMP.MessageType.Step3,
                    minimum: Snake.Crypto.SMP.MessageType.Step3
                  },
                  aToB: {
                    type: "object",
                    properties: {
                      p: {type: "object", format: "binary"},
                      q: {type: "object", format: "binary"},
                      r: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["p", "q", "r"]
                  },
                  bToA: {
                    type: "object",
                    properties: {
                      p: {type: "object", format: "binary"},
                      q: {type: "object", format: "binary"},
                      r: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["p", "q", "r"]
                  }
                },
                additionalProperties: false,
                required: ["type"]
              },
              // Message 4
              {
                type: "object",
                properties: {
                  type: {
                    type: "integer",
                    maximum: Snake.Crypto.SMP.MessageType.Step4,
                    minimum: Snake.Crypto.SMP.MessageType.Step4
                  },
                  aToB: {
                    type: "object",
                    properties: {
                      r: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["r"]
                  },
                  bToA: {
                    type: "object",
                    properties: {
                      r: {type: "object", format: "binary"}
                    },
                    additionalProperties: false,
                    required: ["r"]
                  }
                },
                additionalProperties: false,
                required: ["type"]
              }
            ]
          },
          // TODO: once we properly enforce other messages with oneOf, move
          //       one level out. This probably allows type: group-joined with a
          //       distinct definition for the following.
          group: {
            type: "object",
            oneOf: [
              {$ref: "#/definitions/group-joined"},
              {$ref: "#/definitions/group-leaved"},
              {$ref: "#/definitions/member-added"},
              {$ref: "#/definitions/member-removed"},
              {$ref: "#/definitions/group-admin-access"}
            ]
          }
        },
        publicFields: ["address", "date"],
        authenticationOnlyFields: ["fhmqv"],
        required: ["_recipientType", "_recipientId", "address"],
        additionalProperties: false,
        definitions: {
          bushKey: {
            type: "object",
            properties: {
              key: {type: "object", format: "binary"},
              id: {type: "integer"}
            },
            additionalProperties: false,
            required: ["id", "key"]
          },
          groupKeys: {
            type: "object",
            properties: {
              rootKey: {type: "object", $ref: "#/definitions/bushKey"},
              auxiliaryKeys: {
                type: "array",
                items: {type: "object", $ref: "#/definitions/bushKey"}
              },
              lastValidMessageId: {type: "integer"}
            },
            additionalProperties: false,
            required: ["rootKey", "auxiliaryKeys"]
          },
          nodeRekey: {
            type: "object",
            properties: {
              data: {type: "object", format: "binary"},
              encKey: {type: "integer"},
              withKey: {type: "integer"}
            },
            additionalProperties: false,
            required: ["data", "encKey", "withKey"]
          },
          rekeyMessage: {
            type: "object",
            properties: {
              rootKey: {
                type: "array",
                items: {type: "object", $ref: "#/definitions/nodeRekey"}
              },
              auxiliaryKeys: {
                type: "array",
                items: {type: "object", $ref: "#/definitions/nodeRekey"}
              }
            },
            additionalProperties: false,
            required: ["rootKey", "auxiliaryKeys"]
          },
          "group-joined": {
            type: "object",
            properties: {
              addedBy: {type: "integer"},
              address: {type: "object", format: "address"},
              groupId: {type: "integer"},
              keys: {type: "object", $ref: "#/definitions/groupKeys"},
              lastValidMessageId: {type: "integer"}
            },
            additionalProperties: false,
            required: ["addedBy", "keys", "address"]
          },
          "group-leaved": {
            type: "object",
            properties: {
              removedBy: {type: "integer"},
              groupId: {type: "integer"}
            },
            additionalProperties: false,
            required: ["removedBy"]
          },
          "member-added": {
            type: "object",
            properties: {
              addedBy: {type: "integer"},
              newMember: {type: "integer"},
              // TODO
              rekey: {type: "object"}
            },
            additionalProperties: false,
            required: ["addedBy", "newMember", "rekey"]
          },
          "member-removed": {
            type: "object",
            properties: {
              removedBy: {type: "integer"},
              oldMember: {type: "integer"},
              // TODO
              rekey: {type: "object"}
            },
            additionalProperties: false,
            required: ["removedBy", "oldMember", "rekey"]
          },
          "group-admin-access": {
            type: "object",
            properties: {
              groupId: {type: "integer"},
              promotedBy: {type: "integer"},
              signingPrivateKey: {type: "object", format: "binary"},
              adminKey: {type: "object", format: "binary"},
            },
            additionalProperties: false,
            required: ["promotedBy", "signingPrivateKey", "adminKey"]
          }
        }
      }),
      urlRootToSchemaName: function(urlRoot) {
        return this.mapper[urlRoot];
      },
      getSchemaByUrlRoot: function(urlRoot) {
        return this[this.urlRootToSchemaName(urlRoot)];
      },
      mapper: {
        "username": "user",
        "session-user": "session-user",
        "user": "user",
        "profile": "profile",
        "group": "group",
        "group-admin": "group-admin",
        "message": "message",
        "post": "message",
        "user-history": "user",
        "profile-history": "profile",
        "group-history": "group",
        "message-history": "message",
        "message-data": "message",
        "last-messages": "message",
        "messages-after-id": "message",
        "wot-entry": "wot-entry",
        "common-groups": "wot-entry",
        "wot-entries": "wot-entry"
      }
    });

    // TODO: this is an hack, differentiate the schema for user and session-user
    //       and for post and message
    Snake.Schema["session-user"] = Snake.Schema.user;
    Snake.Schema.post = Snake.Schema.message;

    // Clone the schema removing all the checks on local fields
    Snake.OutgoingSchema = _(Snake.Schema).deepClone();
    localNotRequired(Snake.OutgoingSchema);

  }; // End of the init function
}());
