(function() {
  "use strict";

  // TODO: update code using Promise.every in favor of Promise.all and remove
  // this
  Promise.every = function () {
    return Promise.all(Array.prototype.slice.call(arguments));
  };

  /**
   * Main namespace of the application
   * @namespace Snake
   */
  Snake = _.extend(Snake || {}, {
    /**
     * Clones a template, removes the template class, sets the IDs and
     * renders it with Transparency
     * @memberOf Snake#
     * @param {String} name Name of the template
     * @param {Object} data JSON object used to populate the template with
     * Transparency
     * @param {String} [newId] ID to assign to the new template instance
     * @returns {HTMLElement} Instance of the template
     */
    instanceTemplate: function(name, data, newId, directives) {
      var newInstance = $("#" + name + "-template").clone();
      if (_.isUndefined(newId) || newId.length === 0) {
        newInstance.removeAttr("id");
      } else {
        newInstance.attr("id", newId);
      }
      newInstance.render(data, directives);

      return newInstance;
    },
    /**
     * Formats a date.
     * @memberOf Snake#
     * @param {Date} date the date to format.
     * @returns {String} the formatted date.
     */
    formatDate: function(date) {
      // TODO: do more sophisticated and tunable date formatting or
      // consider to include Moment.js
      if (_.isString(date)) {
        date = new Date(date);
      }

      return date.toString();
    },
    dateComparator: function(direction, attribute) {
      var multiplier;

      direction = direction || "descending";
      attribute = attribute || "date";

      multiplier = (direction === "descending" ? -1 : 1);

      return function(model) {
        // Transform in date, then in number and change sign
        // We do this to reverse the order
        return multiplier * new Date(model.get(attribute));
      };
    },
    errorHandler: function(e) {
      var message;

      message = e;
      if (!_(e.fileName).isUndefined()) {
        message = "Error in " + e.fileName + ":" + e.lineNumber +
            "\n" + e.name + "(" + String(e.number) + "): " + e.message +
            "\n" + e.stack;
      } else if (!_(e.stack).isUndefined()) {
        message = e.toString() + "\n" + e.stack;
      } else if(e.target && e.target.result) {
        message = e.target.result;
      } else if (!_(e.toString).isUndefined()) {
        message = e.toString();
      }

      Snake.Logging.error(this, message);

      throw e;
    },
    identityPromise: function(result) {
      return new Promise(function(resolve, reject) {
        resolve(result);
      });
    },
    slowPokePromise: function(delay) {
      return new Promise(function(resolve, reject) {
        setTimeout(function () {
          resolve();
        }, delay);
      });
    },
    /**
     * Evaluate sequentially an array of functions
     *
     * @param {Function[]} functions: an array of functions to be evaluated.
     *        They may return a Promise.
     * @returns {Promise<Boolean>} a Promise that will evaluate to true if
     *          atleast one of the input functions evaluates to true. Otherwise,
     *          it returns false.
     * @memberOf Snake#
     */
    firstPromise: function(functions) {
      function next(i) {
        return function (res) {
          if (!res) {
            if (i >= functions.length)
              return false;
            return Snake.identityPromise().then(functions[i]).then(next(i+1));
          } else
            return res;
        };
      }

      return Snake.identityPromise().then(functions[0]).then(next(1));
    },
    concatBuffers: function() {
      var result, position, buffers;

      buffers = Array.prototype.slice.call(arguments, 0);

      // If it's not an ArrayBufferView take the underlying ArrayBuffer
      buffers = buffers.map(function(buffer) {
        if ("buffer" in buffer) return buffer.buffer;
        return buffer;
      });

      result = new Uint8Array(buffers.reduce(function(prev, cur) {
        return prev + cur.byteLength;
      }, 0));

      position = 0;
      buffers.forEach(function(buffer) {
        result.set(new Uint8Array(buffer), position);
        position += buffer.byteLength;
      });

      return result;
    },
    generateAddress: function () {
      return window.crypto.getRandomValues(new Uint8Array(128 / 8));
    },
    /**
     * A PromiseQueue is a Promise FIFO queue. When you enqueue, you get back a
     * Promise to wait on for your turn and a function to call when you're done.
     *
     * @return {Object} an object, with two fields: `wait`, the Promise used
     *                  for waiting for your turn, and `unlock` a function to
     *                  call to leave the queue once you're done.
     */
    PromiseQueue: function () {
      return {
        tail: Snake.identityPromise(),
        enqueue: function () {
          var unlock;
          var oldTail = this.tail;
          this.tail = new Promise(function (resolve) { unlock = resolve; });
          return { wait: oldTail, unlock: unlock };
        },
      };
    }
  });

  // TODO: unused right now, consider removal
  // Inspired by: https://gist.github.com/k33g/2287018
  Snake.Base = function() {
    this.initialize && this.initialize.apply(this, arguments);
  };
  Snake.Base.extend = Backbone.Model.extend;

}());
