(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * View used to manage all the profiles.
     * @class Snake.View.ManageProfiles
     */
    ManageProfiles: Snake.View.ViewBase.extend({
      nestedViews: ["form", "profiles"],
      events: {
        "click #add-profile": "addNew"
      },
      initialize: function(options) {
        this.profiles = new Snake.View.UpdatingCollection({
          collection: this.model.profiles,
          childViewConstructor: Snake.View.ListProfiles,
          childViewOptions: {template: "list-profiles"}
        });

        this.form = new Snake.View.Form({
          schema: Snake.Schema.profile,
          legend: "Your Profile",
          button: "Save changes"
        });

        var profiles = this.model.profiles;
        profiles.bind("item:edit", this.edit, this);
        profiles.bind("item:increasePriority", this.increasePriority, this);
        profiles.bind("item:decreasePriority", this.decreasePriority, this);
        profiles.bind("item:remove", this.remove, this);

        this.form.bind("form:close", this.hideForm, this);
        this.form.bind("form:success", this.hideForm, this);
      },
      hideForm: function() {
          this.newProfileForm.hide();
          this.addProfileButton.show();
      },
      showForm: function() {
          this.newProfileForm.show();
          this.addProfileButton.hide();
      },
      /**
       * Removes a profile from the sessionUser profiles.
       * @memberOf Snake.View.ManageProfiles#
       * @param {Object} args
       */
      remove: function(args) {
        // the user has only one profile
        if (this.model.profiles.length === 1) {
          this.showAlertError("Last profile, you can't delete it");
          return;
        }

        // id of the profile
        var id = args.id;

        // the profile to delete is also the public profile
        if (id === this.model.get("publicProfile")) {
          this.showAlertError("Public profile, you can't delete it");
          return;
        }

        var profileToDelete = this.model.profiles.get(id);

        // Removes the profile from the profiles collection, this does
        // not automatically removes the profile also from sessionUser
        this.model.profiles.remove(profileToDelete);

        // Removes the profile also from the sessionUser profiles
        var profiles = this.model.get("profiles");
        profiles = _.filter(profiles, function(profile) {
          if (profile.id !== profileToDelete.id) {
            return profile;
          }
        });

        this.model.save({profiles: profiles}).result.then(
          function() {
            this.showAlertSuccess("Profile successfully deleted");
          }.bind(this),
          function() {
            // TODO if there was an error the profile must be added
            //      again into the profiles collection (maybe...)
            this.showAlertError("There was an error, try again later");
          }.bind(this)
        );
      },
      /**
       *
       * @memberOf Snake.View.ManageProfiles#
       * @param {Object} args
       */
      increasePriority: function(args) {
        var id = args.id;
        var profile = this.model.profiles.get(id);
        var indexProfile = this.model.profiles.indexOf(profile);

        if (indexProfile === 0) {
          this.showAlertError("The profile has already the highest priority");
          return;
        }

        var indexOtherProfile = indexProfile - 1;
        var otherProfile = this.model.profiles.at(indexOtherProfile);

        var priority = profile.get("priority");
        var otherPriority = otherProfile.get("priority");

        profile.save({priority: otherPriority});
        otherProfile.save({priority: priority});

        this.model.profiles.sort();

        // TODO handle errors
        this.showAlertSuccess("Profile priority increased succesfully");
      },
      /**
       *
       * @memberOf Snake.View.ManageProfiles#
       * @param {Object} args
       */
      decreasePriority: function(args) {
        var id = args.id;
        var profile = this.model.profiles.get(id);
        var indexProfile = this.model.profiles.indexOf(profile);

        if (indexProfile === (this.model.profiles.length - 1)) {
          this.showAlertError("The profile has already the lowest priority");
          return;
        }

        var indexOtherProfile = indexProfile + 1;
        var otherProfile = this.model.profiles.at(indexOtherProfile);

        var priority = profile.get("priority");
        var otherPriority = otherProfile.get("priority");

        profile.save({priority: otherPriority});
        otherProfile.save({priority: priority});

        this.model.profiles.sort();

        // TODO handle errors
        this.showAlertSuccess("Profile priority decreased succesfully");
      },
      /**
       *
       * @memberOf Snake.View.ManageProfiles#
       * @param {Object} args
       */
      edit: function(args) {
        var id = args.id;
        var profile = this.model.profiles.get(id);

        this.showForm();

        this.form.changeModel(profile);
      },
      /**
       *
       * @memberOf Snake.View.ManageProfiles#
       */
      addNew: function() {
        // TODO the priority of the lowest priority profile may change
        var lowestPriority = _.last(this.model.profiles.models).get("priority");

        // Creates the new empty profile
        var profile = new Snake.Model.Profile({
          userId: this.model.id,
          priority: lowestPriority - 1
        }, {setDefaultValues: true});

        // Creates the key for the new profile
        var profileKey;
        Snake.Crypto.Common.generateSymmAuthKey().then(function(key) {
          profileKey = key;
          profile.encryptionKey = key;

          this.showForm();
          this.form.changeModel(profile);
        }.bind(this));

        // The new profile doesn't have an id until it's saved on the server
        this.listenToOnce(profile, 'sync', function() {
          // Adds the new profile to the "profiles" collection and
          // to the sessionUser
          this.model.profiles.add(profile);
          this.model.get("profiles").push({
            id: profile.id,
            key: profileKey
          });

          // Saves the new model
          this.model.save();
        }, this);
      },
      render: function() {
        var newManageProfile;

        newManageProfile = Snake.instanceTemplate("manage-profiles",
            this.model.attributeTree(), this.el.id);

        this.renderAllNestedViews(newManageProfile);

        this.$el.replaceWith(newManageProfile);
        this.setElement(newManageProfile[0]);

        this.trigger("title", this.title());

        this.alertSuccess = this.$el.find("#manageProfileSuccess");
        this.alertError = this.$el.find("#manageProfileError");

        this.newProfileForm = this.$el.find("#newProfileForm");
        this.addProfileButton = this.$el.find("#add-profile");

        return this;
      },
      title: function() {
        return "Manage your profiles";
      }
    })
  });
}());
