(function() {
  "use strict";

  Snake.View = _.extend(Snake.View || {}, {
    /**
     * Base class for all the views of the system.
     * Allows easy management of nested views and rendering.
     * @class Snake.View.ViewBase
     */
    ViewBase: Backbone.View.extend({
      /**
       * Renders the specified nested model, replacing the element with
       * class htmlClass with the rendering result.
       * @memberof Snake.View.ViewBase#
       * @param {HTMLElement} newElement the element under which the views
       * have to be rendered
       * @param {String} name name of the nested model to render
       * @param {String} [htmlClass] name of the placeholder element, if
       * omitted name is used
       */
      renderNestedView: function(newElement, name, htmlClass) {
        var placeholder;
        htmlClass = htmlClass || name;

        if (!_.isUndefined(this[name])) {
          placeholder = newElement.find("." + htmlClass);
          if (!this[name].rendered) {
            this[name].setElement(placeholder[0]);
            this[name].render();
          } else {
            placeholder.replaceWith(this[name].el);
          }
        }
      },
      /**
       * Renders all the nested views, as listed in the nestedViews field
       * @memberof Snake.View.ViewBase#
       * @param {HTMLElement} newElement the element under which the views
       * have to be rendered
       */
      renderAllNestedViews: function(newElement) {
        _(this.nestedViews).each(function(view) {
          this.renderNestedView(newElement, view);
        }.bind(this));
      },
      /**
       * Sets the current view to listen to the specfied event on the
       * specified models, using the specified handler
       * @memberof Snake.View.ViewBase#
       * @param {Array} models list of models to attach to
       * @param {String} [event=change] name of the event to listen to
       * @param {Function} [handler=this.render] handler of the event
       */
      renderOn: function(models, event, handler) {
        event = event || "change";
        handler = handler || this.render;
        // Listen to the change events of the model, the sender and the
        // recipient
        _(models).each(
            function(listenModel) {
              if (!_.isUndefined(listenModel)) {
                this.listenTo(listenModel, event, handler);
              }
            }.bind(this));
      },
      /**
       * Shows the alert success message.
       * @memberof Snake.View.ViewBase#
       * @param {String} message the message to show.
       */
      showAlertSuccess: function(message) {
        if (this.alertSuccess) {
          this.alertSuccess.show();
          this.alertSuccess.text(message);
        }

        if (this.alertError) {
          this.alertError.hide();
        }
      },
      /**
       * Shows the alert error message.
       * @memberof Snake.View.ViewBase#
       * @param {String} message the message to show.
       */
      showAlertError: function(message) {
        if (this.alertSuccess) {
          this.alertSuccess.hide();
        }

        if (this.alertError) {
          this.alertError.show();
          this.alertError.text(message);
        }
      }
    })
  });
}());
